import pathlib
import time

import pytest

import salt.config
import salt.crypt
import salt.master
import salt.utils.files
import salt.utils.platform
from tests.support.mock import patch
from tests.support.runtests import RUNTIME_VARS

try:
    import pygit2  # pylint: disable=unused-import

    HAS_PYGIT2 = True
except ImportError:
    HAS_PYGIT2 = False


skipif_no_pygit2 = pytest.mark.skipif(not HAS_PYGIT2, reason="Missing pygit2")


@pytest.fixture
def encrypted_requests(tmp_path):
    # To honor the comment on AESFuncs
    (tmp_path / "pki").mkdir()
    return salt.master.AESFuncs(
        opts={
            "pki_dir": str(tmp_path / "pki"),
            "cachedir": str(tmp_path / "cache"),
            "sock_dir": str(tmp_path / "sock_drawer"),
            "conf_file": str(tmp_path / "config.conf"),
            "fileserver_backend": ["local"],
            "master_job_cache": False,
        }
    )


def test_maintenance_duration():
    """
    Validate Maintenance process duration.
    """
    opts = {
        "loop_interval": 10,
        "maintenance_interval": 1,
        "cachedir": "/tmp",
        "sock_dir": "/tmp",
        "maintenance_niceness": 1,
        "key_cache": "sched",
        "conf_file": "",
        "master_job_cache": "",
        "pki_dir": "/tmp",
        "eauth_tokens": "",
    }
    mp = salt.master.Maintenance(opts)
    with patch("salt.utils.verify.check_max_open_files") as check_files, patch.object(
        mp, "handle_key_cache"
    ) as handle_key_cache, patch("salt.daemons") as salt_daemons, patch.object(
        mp, "handle_git_pillar"
    ) as handle_git_pillar:
        mp.run()
    assert salt_daemons.masterapi.clean_old_jobs.called
    assert salt_daemons.masterapi.clean_expired_tokens.called
    assert salt_daemons.masterapi.clean_pub_auth.called
    assert handle_git_pillar.called


def test_fileserver_duration():
    """
    Validate Fileserver process duration.
    """
    with patch("salt.master.FileserverUpdate._do_update") as update:
        start = time.time()
        salt.master.FileserverUpdate.update(1, {}, 1)
        end = time.time()
        # Interval is equal to timeout so the _do_update method will be called
        # one time.
        update.assert_called_once()
        # Timeout is 1 second
        duration = end - start
        if duration > 2 and salt.utils.platform.spawning_platform():
            # Give spawning platforms some slack
            duration = round(duration, 1)
        assert 2 > duration > 1


@pytest.mark.parametrize(
    "expected_return, payload",
    (
        (
            {
                "jid": "20221107162714826470",
                "id": "example-minion",
                "return": {
                    "pkg_|-linux-install-utils_|-curl_|-installed": {
                        "name": "curl",
                        "changes": {},
                        "result": True,
                        "comment": "All specified packages are already installed",
                        "__sls__": "base-linux.base",
                        "__run_num__": 0,
                        "start_time": "08:27:17.594038",
                        "duration": 32.963,
                        "__id__": "linux-install-utils",
                    },
                },
                "retcode": 0,
                "success": True,
                "fun_args": ["base-linux", {"pillar": {"test": "value"}}],
                "fun": "state.sls",
                "out": "highstate",
            },
            {
                "cmd": "_syndic_return",
                "load": [
                    {
                        "id": "aws.us-east-1.salt-syndic",
                        "jid": "20221107162714826470",
                        "fun": "state.sls",
                        "arg": None,
                        "tgt": None,
                        "tgt_type": None,
                        "load": {
                            "arg": [
                                "base-linux",
                                {"pillar": {"test": "value"}, "__kwarg__": True},
                            ],
                            "cmd": "publish",
                            "fun": "state.sls",
                            "jid": "20221107162714826470",
                            "ret": "",
                            "tgt": "example-minion",
                            "user": "sudo_ubuntu",
                            "kwargs": {
                                "show_jid": False,
                                "delimiter": ":",
                                "show_timeout": True,
                            },
                            "tgt_type": "glob",
                        },
                        "return": {
                            "example-minion": {
                                "return": {
                                    "pkg_|-linux-install-utils_|-curl_|-installed": {
                                        "name": "curl",
                                        "changes": {},
                                        "result": True,
                                        "comment": "All specified packages are already installed",
                                        "__sls__": "base-linux.base",
                                        "__run_num__": 0,
                                        "start_time": "08:27:17.594038",
                                        "duration": 32.963,
                                        "__id__": "linux-install-utils",
                                    },
                                },
                                "retcode": 0,
                                "success": True,
                                "fun_args": [
                                    "base-linux",
                                    {"pillar": {"test": "value"}},
                                ],
                            }
                        },
                        "out": "highstate",
                    }
                ],
                "_stamp": "2022-11-07T16:27:17.965404",
            },
        ),
    ),
)
def test_when_syndic_return_processes_load_then_correct_values_should_be_returned(
    expected_return, payload, encrypted_requests
):
    with patch.object(encrypted_requests, "_return", autospec=True) as fake_return:
        encrypted_requests._syndic_return(payload)
        fake_return.assert_called_with(expected_return)


def test_syndic_return_cache_dir_creation(encrypted_requests):
    """master's cachedir for a syndic will be created by AESFuncs._syndic_return method"""
    cachedir = pathlib.Path(encrypted_requests.opts["cachedir"])
    assert not (cachedir / "syndics").exists()
    encrypted_requests._syndic_return(
        {
            "id": "mamajama",
            "jid": "",
            "return": {},
        }
    )
    assert (cachedir / "syndics").exists()
    assert (cachedir / "syndics" / "mamajama").exists()


def test_syndic_return_cache_dir_creation_traversal(encrypted_requests):
    """
    master's  AESFuncs._syndic_return method cachdir creation is not vulnerable to a directory traversal
    """
    cachedir = pathlib.Path(encrypted_requests.opts["cachedir"])
    assert not (cachedir / "syndics").exists()
    encrypted_requests._syndic_return(
        {
            "id": "../mamajama",
            "jid": "",
            "return": {},
        }
    )
    assert not (cachedir / "syndics").exists()
    assert not (cachedir / "mamajama").exists()


def test_pub_ret_traversal(encrypted_requests, tmp_path):
    """
    master's  AESFuncs._syndic_return method cachdir creation is not vulnerable to a directory traversal
    """
    salt.crypt.gen_keys(tmp_path, "minion", 2048)

    minions = pathlib.Path(encrypted_requests.opts["pki_dir"]) / "minions"
    minions.mkdir()

    with salt.utils.files.fopen(minions / "minion", "wb") as wfp:
        with salt.utils.files.fopen(tmp_path / "minion.pub", "rb") as rfp:
            wfp.write(rfp.read())

    priv = salt.crypt.PrivateKey(tmp_path / "minion.pem")
    with pytest.raises(salt.exceptions.SaltValidationError):
        encrypted_requests.pub_ret(
            {
                "tok": priv.encrypt(b"salt"),
                "id": "minion",
                "jid": "asdf/../../../sdf",
                "return": {},
            }
        )


def _git_pillar_base_config(tmp_path):
    return {
        "__role": "master",
        "pki_dir": str(tmp_path / "pki"),
        "cachedir": str(tmp_path / "cache"),
        "sock_dir": str(tmp_path / "sock_drawer"),
        "conf_file": str(tmp_path / "config.conf"),
        "fileserver_backend": ["local"],
        "master_job_cache": False,
        "file_client": "local",
        "pillar_cache": False,
        "state_top": "top.sls",
        "pillar_roots": {
            "base": [str(tmp_path / "pillar")],
        },
        "render_dirs": [str(pathlib.Path(RUNTIME_VARS.SALT_CODE_DIR) / "renderer")],
        "renderer": "jinja|yaml",
        "renderer_blacklist": [],
        "renderer_whitelist": [],
        "optimization_order": [0, 1, 2],
        "on_demand_ext_pillar": [],
        "git_pillar_user": "",
        "git_pillar_password": "",
        "git_pillar_pubkey": "",
        "git_pillar_privkey": "",
        "git_pillar_passphrase": "",
        "git_pillar_insecure_auth": False,
        "git_pillar_refspecs": salt.config._DFLT_REFSPECS,
        "git_pillar_ssl_verify": True,
        "git_pillar_branch": "master",
        "git_pillar_base": "master",
        "git_pillar_root": "",
        "git_pillar_env": "",
        "git_pillar_fallback": "",
    }


@pytest.fixture
def allowed_funcs(tmp_path):
    """
    Configuration with git on demand pillar allowed
    """
    opts = _git_pillar_base_config(tmp_path)
    opts["on_demand_ext_pillar"] = ["git"]
    salt.crypt.gen_keys(str(tmp_path), "minion", 2048)
    master_pki = tmp_path / "pki"
    master_pki.mkdir()
    accepted_pki = master_pki / "minions"
    accepted_pki.mkdir()
    (accepted_pki / "minion.pub").write_text((tmp_path / "minion.pub").read_text())

    return salt.master.AESFuncs(opts=opts)


@skipif_no_pygit2
def test_on_demand_allowed_command_injection(allowed_funcs, tmp_path, caplog):
    """
    Verify on demand pillars validate remote urls
    """
    pwnpath = tmp_path / "pwn"
    assert not pwnpath.exists()
    load = {
        "cmd": "_pillar",
        "saltenv": "base",
        "pillarenv": "base",
        "id": "carbon",
        "grains": {},
        "ver": 2,
        "ext": {
            "git": [
                f'base ssh://fake@git/repo\n[core]\nsshCommand = touch {pwnpath}\n[remote "origin"]\n'
            ]
        },
        "clean_cache": True,
    }
    with caplog.at_level(level="WARNING"):
        ret = allowed_funcs._pillar(load)
    assert not pwnpath.exists()
    assert "Found bad url data" in caplog.text


@pytest.fixture
def not_allowed_funcs(tmp_path):
    """
    Configuration with no on demand pillars allowed
    """
    opts = _git_pillar_base_config(tmp_path)
    opts["on_demand_ext_pillar"] = []
    salt.crypt.gen_keys(str(tmp_path), "minion", 2048)
    master_pki = tmp_path / "pki"
    master_pki.mkdir()
    accepted_pki = master_pki / "minions"
    accepted_pki.mkdir()
    (accepted_pki / "minion.pub").write_text((tmp_path / "minion.pub").read_text())

    return salt.master.AESFuncs(opts=opts)


def test_on_demand_not_allowed(not_allowed_funcs, tmp_path, caplog):
    """
    Verify on demand pillars do not render when not allowed
    """
    pwnpath = tmp_path / "pwn"
    assert not pwnpath.exists()
    load = {
        "cmd": "_pillar",
        "saltenv": "base",
        "pillarenv": "base",
        "id": "carbon",
        "grains": {},
        "ver": 2,
        "ext": {
            "git": [
                f'base ssh://fake@git/repo\n[core]\nsshCommand = touch {pwnpath}\n[remote "origin"]\n'
            ]
        },
        "clean_cache": True,
    }
    with caplog.at_level(level="WARNING"):
        ret = not_allowed_funcs._pillar(load)
    assert not pwnpath.exists()
    assert (
        "The following ext_pillar modules are not allowed for on-demand pillar data: git."
        in caplog.text
    )



import logging

import pytest

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_if_binaries_missing("ipset", check_all=False),
]


log = logging.getLogger(__name__)


@pytest.fixture(scope="module")
def ipset(modules):
    return modules.ipset


@pytest.fixture()
def setup_set(ipset):
    set_name = "test_name"
    kwargs = {"range": "192.168.0.0/16", "comment": "Hello18"}
    ipset.new_set(name=set_name, set_type="bitmap:ip", family="ipv4", **kwargs)
    yield set_name
    ipset.delete_set(set_name)


def test_ipset_add(ipset, setup_set):
    """
    test ipset.add
    """
    # add set first
    ret = ipset.add(name=setup_set, entry="192.168.0.3 comment Hello18")
    assert ret == "Success"
    check_set = ipset.list_sets()
    assert any([x for x in check_set if x["Name"] == setup_set])


def test_ipset_add_comment_kwarg(ipset, setup_set):
    """
    test ipset.add when comment is set in kwarg
    """
    # add set first
    kwargs = {"comment": "Hello19"}
    entry = "192.168.0.3"
    ret = ipset.add(name=setup_set, entry="192.168.0.3", **kwargs)
    assert ret == "Success"
    check_set = ipset.list_sets()
    assert any([x for x in check_set if x["Name"] == setup_set])



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.states.mysql_grants as mysql_grants
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {mysql_grants: {}}


def test_present():
    """
    Test to ensure that the grant is present with the specified properties.
    """
    name = "frank_exampledb"

    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    mock = MagicMock(side_effect=[True, False, False, False])
    mock_t = MagicMock(return_value=True)
    mock_str = MagicMock(return_value="salt")
    mock_none = MagicMock(return_value=None)
    with patch.dict(
        mysql_grants.__salt__,
        {"mysql.grant_exists": mock, "mysql.grant_add": mock_t},
    ):
        comt = "Grant None on None to None@localhost is already present"
        ret.update({"comment": comt})
        assert mysql_grants.present(name) == ret

        with patch.object(mysql_grants, "_get_mysql_error", mock_str):
            ret.update({"comment": "salt", "result": False})
            assert mysql_grants.present(name) == ret

        with patch.object(mysql_grants, "_get_mysql_error", mock_none):
            with patch.dict(mysql_grants.__opts__, {"test": True}):
                comt = "MySQL grant frank_exampledb is set to be created"
                ret.update({"comment": comt, "result": None})
                assert mysql_grants.present(name) == ret

            with patch.dict(mysql_grants.__opts__, {"test": False}):
                comt = "Grant None on None to None@localhost has been added"
                ret.update(
                    {"comment": comt, "result": True, "changes": {name: "Present"}}
                )
                assert mysql_grants.present(name) == ret


def test_absent():
    """
    Test to ensure that the grant is absent.
    """
    name = "frank_exampledb"

    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    mock = MagicMock(side_effect=[True, False])
    mock_t = MagicMock(side_effect=[True, True, True, False, False])
    mock_str = MagicMock(return_value="salt")
    mock_none = MagicMock(return_value=None)
    with patch.dict(
        mysql_grants.__salt__,
        {"mysql.grant_exists": mock_t, "mysql.grant_revoke": mock},
    ):
        with patch.dict(mysql_grants.__opts__, {"test": True}):
            comt = "MySQL grant frank_exampledb is set to be revoked"
            ret.update({"comment": comt, "result": None})
            assert mysql_grants.absent(name) == ret

        with patch.dict(mysql_grants.__opts__, {"test": False}):
            comt = "Grant None on None for None@localhost has been revoked"
            ret.update({"comment": comt, "result": True, "changes": {name: "Absent"}})
            assert mysql_grants.absent(name) == ret

            with patch.object(mysql_grants, "_get_mysql_error", mock_str):
                comt = "Unable to revoke grant None on None for None@localhost (salt)"
                ret.update({"comment": comt, "result": False, "changes": {}})
                assert mysql_grants.absent(name) == ret

                comt = (
                    "Unable to determine if grant None on "
                    "None for None@localhost exists (salt)"
                )
                ret.update({"comment": comt})
                assert mysql_grants.absent(name) == ret

        with patch.object(mysql_grants, "_get_mysql_error", mock_none):
            comt = (
                "Grant None on None to None@localhost is not present,"
                " so it cannot be revoked"
            )
            ret.update({"comment": comt, "result": True})
            assert mysql_grants.absent(name) == ret



import pytest

pytestmark = [
    pytest.mark.skip_on_windows(reason="Not supported on Windows"),
    pytest.mark.skip_if_binaries_missing(
        "ansible",
        "ansible-doc",
        "ansible-playbook",
        check_all=True,
        reason="ansible is not installed",
    ),
    pytest.mark.slow_test,
    pytest.mark.timeout_unless_on_windows(120),
]


@pytest.fixture
def ansible(modules):
    return modules.ansible


def test_short_alias(modules, ansible):
    """
    Test that the ansible functions are actually loaded and we can target using the short alias.
    """
    ret = ansible.ping()
    assert ret == {"ping": "pong"}

    ansible_ping_func = None
    if "ansible.system.ping" in modules:
        # we need to go by getattr() because salt's loader will try to find "system" in the dictionary and fail
        # The ansible hack injects, in this case, "system.ping" as an attribute to the loaded module
        ansible_ping_func = getattr(modules.ansible, "system.ping")
        # Make sure we don't set the full ansible module
        assert "ansible.ansible.system.ping" not in modules
    elif "ansible.builtin.ping" in modules:
        # Ansible >= 2.14
        # we need to go by getattr() because salt's loader will try to find "builtin" in the dictionary and fail
        # The ansible hack injects, in this case, "builtin.ping" as an attribute to the loaded module
        ansible_ping_func = getattr(modules.ansible, "builtin.ping")
        # Make sure we don't set the full ansible module
        assert "ansible.ansible.builtin.ping" not in modules

    # if ansible_ping_func is None, then it's ok, we already ran 'ping' at the
    # top if this test, so it worked.
    if ansible_ping_func:
        ret = ansible_ping_func()
        assert ret == {"ping": "pong"}


def test_passing_data_to_ansible_modules(ansible):
    """
    Test that the ansible functions are actually loaded
    """
    expected = "foobar"
    ret = ansible.ping(data=expected)
    assert ret == {"ping": expected}



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import salt.modules.neutron as neutron
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock
from tests.support.unit import TestCase


class MockNeutron:
    """
    Mock of neutron
    """

    @staticmethod
    def get_quotas_tenant():
        """
        Mock of get_quotas_tenant method
        """
        return True

    @staticmethod
    def list_quotas():
        """
        Mock of list_quotas method
        """
        return True

    @staticmethod
    def show_quota(tenant_id):
        """
        Mock of show_quota method
        """
        return tenant_id

    @staticmethod
    def update_quota(
        tenant_id,
        subnet,
        router,
        network,
        floatingip,
        port,
        security_group,
        security_group_rule,
    ):
        """
        Mock of update_quota method
        """
        return (
            tenant_id,
            subnet,
            router,
            network,
            floatingip,
            port,
            security_group,
            security_group_rule,
        )

    @staticmethod
    def delete_quota(tenant_id):
        """
        Mock of delete_quota method
        """
        return tenant_id

    @staticmethod
    def list_extensions():
        """
        Mock of list_extensions method
        """
        return True

    @staticmethod
    def list_ports():
        """
        Mock of list_ports method
        """
        return True

    @staticmethod
    def show_port(port):
        """
        Mock of show_port method
        """
        return port

    @staticmethod
    def create_port(name, network, device_id, admin_state_up):
        """
        Mock of create_port method
        """
        return (name, network, device_id, admin_state_up)

    @staticmethod
    def update_port(port, name, admin_state_up):
        """
        Mock of update_port method
        """
        return (port, name, admin_state_up)

    @staticmethod
    def delete_port(port):
        """
        Mock of delete_port method
        """
        return port

    @staticmethod
    def list_networks():
        """
        Mock of list_networks method
        """
        return True

    @staticmethod
    def show_network(network):
        """
        Mock of show_network method
        """
        return network

    @staticmethod
    def create_network(
        name,
        admin_state_up,
        router_ext,
        network_type,
        physical_network,
        segmentation_id,
        shared,
    ):
        """
        Mock of create_network method
        """
        return (
            name,
            admin_state_up,
            router_ext,
            network_type,
            physical_network,
            segmentation_id,
            shared,
        )

    @staticmethod
    def update_network(network, name):
        """
        Mock of update_network method
        """
        return (network, name)

    @staticmethod
    def delete_network(network):
        """
        Mock of delete_network method
        """
        return network

    @staticmethod
    def list_subnets():
        """
        Mock of list_subnets method
        """
        return True

    @staticmethod
    def show_subnet(subnet):
        """
        Mock of show_subnet method
        """
        return subnet

    @staticmethod
    def create_subnet(network, cidr, name, ip_version):
        """
        Mock of create_subnet method
        """
        return (network, cidr, name, ip_version)

    @staticmethod
    def update_subnet(subnet, name):
        """
        Mock of update_subnet method
        """
        return (subnet, name)

    @staticmethod
    def delete_subnet(subnet):
        """
        Mock of delete_subnet method
        """
        return subnet

    @staticmethod
    def list_routers():
        """
        Mock of list_routers method
        """
        return True

    @staticmethod
    def show_router(router):
        """
        Mock of show_router method
        """
        return router

    @staticmethod
    def create_router(name, ext_network, admin_state_up):
        """
        Mock of create_router method
        """
        return (name, ext_network, admin_state_up)

    @staticmethod
    def update_router(router, name, admin_state_up, **kwargs):
        """
        Mock of update_router method
        """
        return (router, name, admin_state_up, kwargs)

    @staticmethod
    def delete_router(router):
        """
        Mock of delete_router method
        """
        return router

    @staticmethod
    def add_interface_router(router, subnet):
        """
        Mock of add_interface_router method
        """
        return (router, subnet)

    @staticmethod
    def remove_interface_router(router, subnet):
        """
        Mock of remove_interface_router method
        """
        return (router, subnet)

    @staticmethod
    def add_gateway_router(router, ext_network):
        """
        Mock of add_gateway_router method
        """
        return (router, ext_network)

    @staticmethod
    def remove_gateway_router(router):
        """
        Mock of remove_gateway_router method
        """
        return router

    @staticmethod
    def list_floatingips():
        """
        Mock of list_floatingips method
        """
        return True

    @staticmethod
    def show_floatingip(floatingip_id):
        """
        Mock of show_floatingip method
        """
        return floatingip_id

    @staticmethod
    def create_floatingip(floating_network, port):
        """
        Mock of create_floatingip method
        """
        return (floating_network, port)

    @staticmethod
    def update_floatingip(floating_network, port):
        """
        Mock of create_floatingip method
        """
        return (floating_network, port)

    @staticmethod
    def delete_floatingip(floatingip_id):
        """
        Mock of delete_floatingip method
        """
        return floatingip_id

    @staticmethod
    def list_security_groups():
        """
        Mock of list_security_groups method
        """
        return True

    @staticmethod
    def show_security_group(security_group):
        """
        Mock of show_security_group method
        """
        return security_group

    @staticmethod
    def create_security_group(name, description):
        """
        Mock of create_security_group method
        """
        return (name, description)

    @staticmethod
    def update_security_group(security_group, name, description):
        """
        Mock of update_security_group method
        """
        return (security_group, name, description)

    @staticmethod
    def delete_security_group(security_group):
        """
        Mock of delete_security_group method
        """
        return security_group

    @staticmethod
    def list_security_group_rules():
        """
        Mock of list_security_group_rules method
        """
        return True

    @staticmethod
    def show_security_group_rule(security_group_rule_id):
        """
        Mock of show_security_group_rule method
        """
        return security_group_rule_id

    @staticmethod
    def create_security_group_rule(
        security_group,
        remote_group_id,
        direction,
        protocol,
        port_range_min,
        port_range_max,
        ethertype,
    ):
        """
        Mock of create_security_group_rule method
        """
        return (
            security_group,
            remote_group_id,
            direction,
            protocol,
            port_range_min,
            port_range_max,
            ethertype,
        )

    @staticmethod
    def delete_security_group_rule(security_group_rule_id):
        """
        Mock of delete_security_group_rule method
        """
        return security_group_rule_id

    @staticmethod
    def list_vpnservices(retrieve_all, **kwargs):
        """
        Mock of list_vpnservices method
        """
        return (retrieve_all, kwargs)

    @staticmethod
    def show_vpnservice(vpnservice, **kwargs):
        """
        Mock of show_vpnservice method
        """
        return (vpnservice, kwargs)

    @staticmethod
    def create_vpnservice(subnet, router, name, admin_state_up):
        """
        Mock of create_vpnservice method
        """
        return (subnet, router, name, admin_state_up)

    @staticmethod
    def update_vpnservice(vpnservice, desc):
        """
        Mock of update_vpnservice method
        """
        return (vpnservice, desc)

    @staticmethod
    def delete_vpnservice(vpnservice):
        """
        Mock of delete_vpnservice method
        """
        return vpnservice

    @staticmethod
    def list_ipsec_site_connections():
        """
        Mock of list_ipsec_site_connections method
        """
        return True

    @staticmethod
    def show_ipsec_site_connection(ipsec_site_connection):
        """
        Mock of show_ipsec_site_connection method
        """
        return ipsec_site_connection

    @staticmethod
    def create_ipsec_site_connection(
        name,
        ipsecpolicy,
        ikepolicy,
        vpnservice,
        peer_cidrs,
        peer_address,
        peer_id,
        psk,
        admin_state_up,
        **kwargs
    ):
        """
        Mock of create_ipsec_site_connection method
        """
        return (
            name,
            ipsecpolicy,
            ikepolicy,
            vpnservice,
            peer_cidrs,
            peer_address,
            peer_id,
            psk,
            admin_state_up,
            kwargs,
        )

    @staticmethod
    def delete_ipsec_site_connection(ipsec_site_connection):
        """
        Mock of delete_vpnservice method
        """
        return ipsec_site_connection

    @staticmethod
    def list_ikepolicies():
        """
        Mock of list_ikepolicies method
        """
        return True

    @staticmethod
    def show_ikepolicy(ikepolicy):
        """
        Mock of show_ikepolicy method
        """
        return ikepolicy

    @staticmethod
    def create_ikepolicy(name, **kwargs):
        """
        Mock of create_ikepolicy method
        """
        return (name, kwargs)

    @staticmethod
    def delete_ikepolicy(ikepolicy):
        """
        Mock of delete_ikepolicy method
        """
        return ikepolicy

    @staticmethod
    def list_ipsecpolicies():
        """
        Mock of list_ipsecpolicies method
        """
        return True

    @staticmethod
    def show_ipsecpolicy(ipsecpolicy):
        """
        Mock of show_ipsecpolicy method
        """
        return ipsecpolicy

    @staticmethod
    def create_ipsecpolicy(name, **kwargs):
        """
        Mock of create_ikepolicy method
        """
        return (name, kwargs)

    @staticmethod
    def delete_ipsecpolicy(ipsecpolicy):
        """
        Mock of delete_ipsecpolicy method
        """
        return ipsecpolicy


class NeutronTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.neutron
    """

    def setup_loader_modules(self):
        return {neutron: {"_auth": MagicMock(return_value=MockNeutron())}}

    # 'get_quotas_tenant' function tests: 1

    def test_get_quotas_tenant(self):
        """
        Test if it fetches tenant info in server's context for
        following quota operation
        """
        self.assertTrue(neutron.get_quotas_tenant(profile="openstack1"))

    # 'list_quotas' function tests: 1

    def test_list_quotas(self):
        """
        Test if it fetches all tenants quotas
        """
        self.assertTrue(neutron.list_quotas(profile="openstack1"))

    # 'show_quota' function tests: 1

    def test_show_quota(self):
        """
        Test if it fetches information of a certain tenant's quotas
        """
        self.assertTrue(neutron.show_quota("Salt", profile="openstack1"))

    # 'update_quota' function tests: 1

    def test_update_quota(self):
        """
        Test if it update a tenant's quota
        """
        self.assertTrue(
            neutron.update_quota(
                "Salt",
                subnet="40",
                router="50",
                network="10",
                floatingip="30",
                port="30",
                security_group="10",
                security_group_rule="SS",
            )
        )

    # 'delete_quota' function tests: 1

    def test_delete_quota(self):
        """
        Test if it delete the specified tenant's quota value
        """
        self.assertTrue(neutron.delete_quota("Salt", profile="openstack1"))

    # 'list_extensions' function tests: 1

    def test_list_extensions(self):
        """
        Test if it fetches a list of all extensions on server side
        """
        self.assertTrue(neutron.list_extensions(profile="openstack1"))

    # 'list_ports' function tests: 1

    def test_list_ports(self):
        """
        Test if it fetches a list of all networks for a tenant
        """
        self.assertTrue(neutron.list_ports(profile="openstack1"))

    # 'show_port' function tests: 1

    def test_show_port(self):
        """
        Test if it fetches information of a certain port
        """
        self.assertTrue(neutron.show_port("1080", profile="openstack1"))

    # 'create_port' function tests: 1

    def test_create_port(self):
        """
        Test if it creates a new port
        """
        self.assertTrue(
            neutron.create_port(
                "Salt",
                "SALTSTACK",
                device_id="800",
                admin_state_up=True,
                profile="openstack1",
            )
        )

    # 'update_port' function tests: 1

    def test_update_port(self):
        """
        Test if it updates a port
        """
        self.assertTrue(
            neutron.update_port(
                "800", "SALTSTACK", admin_state_up=True, profile="openstack1"
            )
        )

    # 'delete_port' function tests: 1

    def test_delete_port(self):
        """
        Test if it deletes the specified port
        """
        self.assertTrue(neutron.delete_port("1080", profile="openstack1"))

    # 'list_networks' function tests: 1

    def test_list_networks(self):
        """
        Test if it fetches a list of all networks for a tenant
        """
        self.assertTrue(neutron.list_networks(profile="openstack1"))

    # 'show_network' function tests: 1

    def test_show_network(self):
        """
        Test if it fetches information of a certain network
        """
        self.assertTrue(neutron.show_network("SALTSTACK", profile="openstack1"))

    # 'create_network' function tests: 1

    def test_create_network(self):
        """
        Test if it creates a new network
        """
        self.assertTrue(neutron.create_network("SALT", profile="openstack1"))

    # 'update_network' function tests: 1

    def test_update_network(self):
        """
        Test if it updates a network
        """
        self.assertTrue(
            neutron.update_network("SALT", "SLATSTACK", profile="openstack1")
        )

    # 'delete_network' function tests: 1

    def test_delete_network(self):
        """
        Test if it deletes the specified network
        """
        self.assertTrue(neutron.delete_network("SALTSTACK", profile="openstack1"))

    # 'list_subnets' function tests: 1

    def test_list_subnets(self):
        """
        Test if it fetches a list of all networks for a tenant
        """
        self.assertTrue(neutron.list_subnets(profile="openstack1"))

    # 'show_subnet' function tests: 1

    def test_show_subnet(self):
        """
        Test if it fetches information of a certain subnet
        """
        self.assertTrue(neutron.show_subnet("SALTSTACK", profile="openstack1"))

    # 'create_subnet' function tests: 1

    def test_create_subnet(self):
        """
        Test if it creates a new subnet
        """
        self.assertTrue(
            neutron.create_subnet(
                "192.168.1.0",
                "192.168.1.0/24",
                name="Salt",
                ip_version=4,
                profile="openstack1",
            )
        )

    # 'update_subnet' function tests: 1

    def test_update_subnet(self):
        """
        Test if it updates a subnet
        """
        self.assertTrue(
            neutron.update_subnet("255.255.255.0", name="Salt", profile="openstack1")
        )

    # 'delete_subnet' function tests: 1

    def test_delete_subnet(self):
        """
        Test if it deletes the specified subnet
        """
        self.assertTrue(neutron.delete_subnet("255.255.255.0", profile="openstack1"))

    # 'list_routers' function tests: 1

    def test_list_routers(self):
        """
        Test if it fetches a list of all routers for a tenant
        """
        self.assertTrue(neutron.list_routers(profile="openstack1"))

    # 'show_router' function tests: 1

    def test_show_router(self):
        """
        Test if it fetches information of a certain router
        """
        self.assertTrue(neutron.show_router("SALTSTACK", profile="openstack1"))

    # 'create_router' function tests: 1

    def test_create_router(self):
        """
        Test if it creates a new router
        """
        self.assertTrue(
            neutron.create_router(
                "SALT", "192.168.1.0", admin_state_up=True, profile="openstack1"
            )
        )

    # 'update_router' function tests: 1

    def test_update_router(self):
        """
        Test if it updates a router
        """
        self.assertTrue(
            neutron.update_router("255.255.255.0", name="Salt", profile="openstack1")
        )

    # 'delete_router' function tests: 1

    def test_delete_router(self):
        """
        Test if it delete the specified router
        """
        self.assertTrue(neutron.delete_router("SALTSTACK", profile="openstack1"))

    # 'add_interface_router' function tests: 1

    def test_add_interface_router(self):
        """
        Test if it adds an internal network interface to the specified router
        """
        self.assertTrue(
            neutron.add_interface_router("Salt", "255.255.255.0", profile="openstack1")
        )

    # 'remove_interface_router' function tests: 1

    def test_remove_interface_router(self):
        """
        Test if it removes an internal network interface from the specified
        router
        """
        self.assertTrue(
            neutron.remove_interface_router(
                "Salt", "255.255.255.0", profile="openstack1"
            )
        )

    # 'add_gateway_router' function tests: 1

    def test_add_gateway_router(self):
        """
        Test if it adds an external network gateway to the specified router
        """
        self.assertTrue(
            neutron.add_gateway_router("Salt", "SALTSTACK", profile="openstack1")
        )

    # 'remove_gateway_router' function tests: 1

    def test_remove_gateway_router(self):
        """
        Test if it removes an external network gateway from the specified router
        """

        self.assertTrue(
            neutron.remove_gateway_router("SALTSTACK", profile="openstack1")
        )

    # 'list_floatingips' function tests: 1

    def test_list_floatingips(self):
        """
        Test if it fetch a list of all floatingIPs for a tenant
        """
        self.assertTrue(neutron.list_floatingips(profile="openstack1"))

    # 'show_floatingip' function tests: 1

    def test_show_floatingip(self):
        """
        Test if it fetches information of a certain floatingIP
        """
        self.assertTrue(neutron.show_floatingip("SALTSTACK", profile="openstack1"))

    # 'create_floatingip' function tests: 1

    def test_create_floatingip(self):
        """
        Test if it creates a new floatingIP
        """
        self.assertTrue(
            neutron.create_floatingip("SALTSTACK", port="800", profile="openstack1")
        )

    # 'update_floatingip' function tests: 1

    def test_update_floatingip(self):
        """
        Test if it updates a floatingIP
        """
        self.assertTrue(
            neutron.update_floatingip("SALTSTACK", port="800", profile="openstack1")
        )

    # 'delete_floatingip' function tests: 1

    def test_delete_floatingip(self):
        """
        Test if it deletes the specified floating IP
        """
        self.assertTrue(neutron.delete_floatingip("SALTSTACK", profile="openstack1"))

    # 'list_security_groups' function tests: 1

    def test_list_security_groups(self):
        """
        Test if it fetches a list of all security groups for a tenant
        """
        self.assertTrue(neutron.list_security_groups(profile="openstack1"))

    # 'show_security_group' function tests: 1

    def test_show_security_group(self):
        """
        Test if it fetches information of a certain security group
        """
        self.assertTrue(neutron.show_security_group("SALTSTACK", profile="openstack1"))

    # 'create_security_group' function tests: 1

    def test_create_security_group(self):
        """
        Test if it creates a new security group
        """
        self.assertTrue(
            neutron.create_security_group(
                "SALTSTACK", "Security group", profile="openstack1"
            )
        )

    # 'update_security_group' function tests: 1

    def test_update_security_group(self):
        """
        Test if it updates a security group
        """
        self.assertTrue(
            neutron.update_security_group(
                "SALT", "SALTSTACK", "Security group", profile="openstack1"
            )
        )

    # 'delete_security_group' function tests: 1

    def test_delete_security_group(self):
        """
        Test if it deletes the specified security group
        """
        self.assertTrue(neutron.delete_security_group("SALT", profile="openstack1"))

    # 'list_security_group_rules' function tests: 1

    def test_list_security_group_rules(self):
        """
        Test if it fetches a list of all security group rules for a tenant
        """
        self.assertTrue(neutron.list_security_group_rules(profile="openstack1"))

    # 'show_security_group_rule' function tests: 1

    def test_show_security_group_rule(self):
        """
        Test if it fetches information of a certain security group rule
        """
        self.assertTrue(
            neutron.show_security_group_rule("SALTSTACK", profile="openstack1")
        )

    # 'create_security_group_rule' function tests: 1

    def test_create_security_group_rule(self):
        """
        Test if it creates a new security group rule
        """
        self.assertTrue(
            neutron.create_security_group_rule("SALTSTACK", profile="openstack1")
        )

    # 'delete_security_group_rule' function tests: 1

    def test_delete_security_group_rule(self):
        """
        Test if it deletes the specified security group rule
        """
        self.assertTrue(
            neutron.delete_security_group_rule("SALTSTACK", profile="openstack1")
        )

    # 'list_vpnservices' function tests: 1

    def test_list_vpnservices(self):
        """
        Test if it fetches a list of all configured VPN services for a tenant
        """
        self.assertTrue(neutron.list_vpnservices(True, profile="openstack1"))

    # 'show_vpnservice' function tests: 1

    def test_show_vpnservice(self):
        """
        Test if it fetches information of a specific VPN service
        """
        self.assertTrue(neutron.show_vpnservice("SALT", profile="openstack1"))

    # 'create_vpnservice' function tests: 1

    def test_create_vpnservice(self):
        """
        Test if it creates a new VPN service
        """
        self.assertTrue(
            neutron.create_vpnservice(
                "255.255.255.0", "SALT", "SALTSTACK", True, profile="openstack1"
            )
        )

    # 'update_vpnservice' function tests: 1

    def test_update_vpnservice(self):
        """
        Test if it updates a VPN service
        """
        self.assertTrue(
            neutron.update_vpnservice("SALT", "VPN Service1", profile="openstack1")
        )

    # 'delete_vpnservice' function tests: 1

    def test_delete_vpnservice(self):
        """
        Test if it deletes the specified VPN service
        """
        self.assertTrue(
            neutron.delete_vpnservice("SALT VPN Service1", profile="openstack1")
        )

    # 'list_ipsec_site_connections' function tests: 1

    def test_list_ipsec_site(self):
        """
        Test if it fetches all configured IPsec Site Connections for a tenant
        """
        self.assertTrue(neutron.list_ipsec_site_connections(profile="openstack1"))

    # 'show_ipsec_site_connection' function tests: 1

    def test_show_ipsec_site_connection(self):
        """
        Test if it fetches information of a specific IPsecSiteConnection
        """
        self.assertTrue(
            neutron.show_ipsec_site_connection("SALT", profile="openstack1")
        )

    # 'create_ipsec_site_connection' function tests: 1

    def test_create_ipsec_site(self):
        """
        Test if it creates a new IPsecSiteConnection
        """
        self.assertTrue(
            neutron.create_ipsec_site_connection(
                "SALTSTACK",
                "A",
                "B",
                "C",
                "192.168.1.0/24",
                "192.168.1.11",
                "192.168.1.10",
                "secret",
                profile="openstack1",
            )
        )

    # 'delete_ipsec_site_connection' function tests: 1

    def test_delete_ipsec_site(self):
        """
        Test if it deletes the specified IPsecSiteConnection
        """
        self.assertTrue(
            neutron.delete_ipsec_site_connection(
                "SALT VPN Service1", profile="openstack1"
            )
        )

    # 'list_ikepolicies' function tests: 1

    def test_list_ikepolicies(self):
        """
        Test if it fetches a list of all configured IKEPolicies for a tenant
        """
        self.assertTrue(neutron.list_ikepolicies(profile="openstack1"))

    # 'show_ikepolicy' function tests: 1

    def test_show_ikepolicy(self):
        """
        Test if it fetches information of a specific IKEPolicy
        """
        self.assertTrue(neutron.show_ikepolicy("SALT", profile="openstack1"))

    # 'create_ikepolicy' function tests: 1

    def test_create_ikepolicy(self):
        """
        Test if it creates a new IKEPolicy
        """
        self.assertTrue(neutron.create_ikepolicy("SALTSTACK", profile="openstack1"))

    # 'delete_ikepolicy' function tests: 1

    def test_delete_ikepolicy(self):
        """
        Test if it deletes the specified IKEPolicy
        """
        self.assertTrue(neutron.delete_ikepolicy("SALT", profile="openstack1"))

    # 'list_ipsecpolicies' function tests: 1

    def test_list_ipsecpolicies(self):
        """
        Test if it fetches a list of all configured IPsecPolicies for a tenant
        """
        self.assertTrue(neutron.list_ipsecpolicies(profile="openstack1"))

    # 'show_ipsecpolicy' function tests: 1

    def test_show_ipsecpolicy(self):
        """
        Test if it fetches information of a specific IPsecPolicy
        """
        self.assertTrue(neutron.show_ipsecpolicy("SALT", profile="openstack1"))

    # 'create_ipsecpolicy' function tests: 1

    def test_create_ipsecpolicy(self):
        """
        Test if it creates a new IPsecPolicy
        """
        self.assertTrue(neutron.create_ipsecpolicy("SALTSTACK", profile="openstack1"))

    # 'delete_ipsecpolicy' function tests: 1

    def test_delete_ipsecpolicy(self):
        """
        Test if it deletes the specified IPsecPolicy
        """
        self.assertTrue(neutron.delete_ipsecpolicy("SALT", profile="openstack1"))



"""
    :codeauthor: Rahul Handay <rahulha@saltstack.com>
"""

import pytest

import salt.states.tomcat as tomcat
from salt.modules import tomcat as tomcatmod
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {tomcat: {"__env__": "base"}}


def test_war_deployed():
    """
    Test to enforce that the WAR will be deployed and
    started in the context path it will make use of WAR versions
    """
    ret = {"name": "salt", "changes": {}, "result": True, "comment": ""}
    mock_start = MagicMock(return_value="saltstack")
    mock_undeploy = MagicMock(side_effect=["FAIL", "saltstack"])
    mock_deploy = MagicMock(return_value="deploy")
    mock_ls = MagicMock(
        side_effect=[
            {"salt": {"version": "jenkins-1.20.4", "mode": "running"}},
            {"salt": {"version": "1"}},
            {"salt": {"version": "jenkins-1.2.4", "mode": "run"}},
            {"salt": {"version": "1"}},
            {"salt": {"version": "1"}},
            {"salt": {"version": "1"}},
        ]
    )
    with patch.dict(
        tomcat.__salt__,
        {
            "tomcat.ls": mock_ls,
            "tomcat.extract_war_version": tomcatmod.extract_war_version,
            "tomcat.start": mock_start,
            "tomcat.undeploy": mock_undeploy,
            "tomcat.deploy_war": mock_deploy,
        },
    ):
        ret.update({"comment": "salt with version 1.20.4 is already deployed"})
        assert tomcat.war_deployed("salt", "salt://jenkins-1.20.4.war") == ret

        with patch.dict(tomcat.__opts__, {"test": True}):
            ret.update(
                {
                    "changes": {
                        "deploy": "will deploy salt with version 1.2.4",
                        "undeploy": "undeployed salt with version 1",
                    },
                    "result": None,
                    "comment": "",
                }
            )
            assert tomcat.war_deployed("salt", "salt://jenkins-1.2.4.war") == ret

        with patch.dict(tomcat.__opts__, {"test": False}):
            ret.update(
                {
                    "changes": {"start": "starting salt"},
                    "comment": "saltstack",
                    "result": False,
                }
            )
            assert tomcat.war_deployed("salt", "salt://jenkins-1.2.4.war") == ret

            ret.update(
                {
                    "changes": {
                        "deploy": "will deploy salt with version 1.2.4",
                        "undeploy": "undeployed salt with version 1",
                    },
                    "comment": "FAIL",
                }
            )
            assert tomcat.war_deployed("salt", "salt://jenkins-1.2.4.war") == ret

            ret.update(
                {
                    "changes": {"undeploy": "undeployed salt with version 1"},
                    "comment": "deploy",
                }
            )
            assert tomcat.war_deployed("salt", "salt://jenkins-1.2.4.war") == ret


def test_war_deployed_no_version():
    """
    Tests that going from versions to no versions and back work, as well
    as not overwriting a WAR without version with another without version.
    """
    ret = {"name": "salt", "changes": {}, "result": None, "comment": ""}

    mock_deploy = MagicMock(return_value="deploy")
    mock_undeploy = MagicMock(return_value="SUCCESS")
    mock_ls_version = MagicMock(
        return_value={"salt": {"version": "1.2.4", "mode": "running"}}
    )
    mock_ls_no_version = MagicMock(
        return_value={"salt": {"version": "", "mode": "running"}}
    )

    with patch.dict(tomcat.__opts__, {"test": True}):
        with patch.dict(
            tomcat.__salt__,
            {
                "tomcat.ls": mock_ls_version,
                "tomcat.extract_war_version": tomcatmod.extract_war_version,
                "tomcat.deploy_war": mock_deploy,
                "tomcat.undeploy": mock_undeploy,
            },
        ):
            ret.update(
                {
                    "changes": {
                        "deploy": "will deploy salt with no version",
                        "undeploy": "undeployed salt with version 1.2.4",
                    },
                }
            )
            assert tomcat.war_deployed("salt", "salt://jenkins.war") == ret

        with patch.dict(
            tomcat.__salt__,
            {
                "tomcat.ls": mock_ls_no_version,
                "tomcat.extract_war_version": tomcatmod.extract_war_version,
                "tomcat.deploy_war": mock_deploy,
                "tomcat.undeploy": mock_undeploy,
            },
        ):
            ret.update(
                {
                    "changes": {
                        "deploy": "will deploy salt with version 1.2.4",
                        "undeploy": "undeployed salt with no version",
                    },
                }
            )
            assert (
                tomcat.war_deployed("salt", "salt://jenkins.war", version="1.2.4")
                == ret
            )
            assert tomcat.war_deployed("salt", "salt://jenkins-1.2.4.war") == ret
            ret.update(
                {
                    "changes": {},
                    "comment": "salt with no version is already deployed",
                    "result": True,
                }
            )
            assert tomcat.war_deployed("salt", "salt://jenkins.war") == ret
            assert (
                tomcat.war_deployed("salt", "salt://jenkins-1.2.4.war", version=False)
                == ret
            )


def test_wait():
    """
    Test to wait for the tomcat manager to load
    """
    ret = {
        "name": "salt",
        "changes": {},
        "result": True,
        "comment": "tomcat manager is ready",
    }
    mock = MagicMock(return_value=True)
    with patch.dict(
        tomcat.__salt__,
        {
            "tomcat.status": mock,
            "tomcat.extract_war_version": tomcatmod.extract_war_version,
        },
    ):
        assert tomcat.wait("salt") == ret


def test_mod_watch():
    """
    Test to the tomcat watcher function.
    """
    ret = {"name": "salt", "changes": {}, "result": False, "comment": "True"}
    mock = MagicMock(return_value="True")
    with patch.dict(
        tomcat.__salt__,
        {
            "tomcat.reload": mock,
            "tomcat.extract_war_version": tomcatmod.extract_war_version,
        },
    ):
        ret.update({"changes": {"salt": False}})
        assert tomcat.mod_watch("salt") == ret


def test_undeployed():
    """
    Test to enforce that the WAR will be un-deployed from the server
    """
    ret = {"name": "salt", "changes": {}, "result": False, "comment": "True"}
    mock = MagicMock(side_effect=[False, True, True, True, True])
    mock1 = MagicMock(
        side_effect=[
            {"salt": {"a": 1}},
            {"salt": {"version": 1}},
            {"salt": {"version": 1}},
            {"salt": {"version": 1}},
        ]
    )
    mock2 = MagicMock(side_effect=["FAIL", "saltstack"])
    with patch.dict(
        tomcat.__salt__,
        {
            "tomcat.status": mock,
            "tomcat.extract_war_version": tomcatmod.extract_war_version,
            "tomcat.ls": mock1,
            "tomcat.undeploy": mock2,
        },
    ):
        ret.update({"comment": "Tomcat Manager does not respond"})
        assert tomcat.undeployed("salt") == ret

        ret.update({"comment": "", "result": True})
        assert tomcat.undeployed("salt") == ret

        with patch.dict(tomcat.__opts__, {"test": True}):
            ret.update({"changes": {"undeploy": 1}, "result": None})
            assert tomcat.undeployed("salt") == ret

        with patch.dict(tomcat.__opts__, {"test": False}):
            ret.update({"changes": {"undeploy": 1}, "comment": "FAIL", "result": False})
            assert tomcat.undeployed("salt") == ret

            ret.update({"changes": {"undeploy": 1}, "comment": "", "result": True})
            assert tomcat.undeployed("salt") == ret



import platform

import pytest
from saltfactories.utils import random_string

import salt.modules.win_useradd
import salt.utils.win_functions as win_functions
from tests.support.mock import MagicMock, patch

HAS_WIN32 = False
HAS_PYWIN = False

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
]

try:
    import win32net

    HAS_WIN32 = True

    class WinError(win32net.error):
        winerror = 0

except ImportError:
    HAS_WIN32 = False

try:
    import pywintypes

    HAS_PYWIN = True

    class PyWinError(pywintypes.error):
        pywinerror = 0

except ImportError:
    HAS_PYWIN = False


@pytest.fixture(scope="module")
def test_user():
    user_name = random_string("test-")
    salt.modules.win_useradd.add(name=user_name, password="P@ssw0rd")
    yield user_name
    salt.modules.win_useradd.delete(name=user_name)


# Test cases for salt.utils.win_functions.


def test_escape_argument_simple():
    """
    Test to make sure we encode simple arguments correctly
    """
    encoded = win_functions.escape_argument("simple")
    assert encoded == "simple"


def test_escape_argument_with_space():
    """
    Test to make sure we encode arguments containing spaces correctly
    """
    encoded = win_functions.escape_argument("with space")
    assert encoded == '^"with space^"'


def test_escape_argument_simple_path():
    """
    Test to make sure we encode simple path arguments correctly
    """
    encoded = win_functions.escape_argument("C:\\some\\path")
    assert encoded == "C:\\some\\path"


def test_escape_argument_path_with_space():
    """
    Test to make sure we encode path arguments containing spaces correctly
    """
    encoded = win_functions.escape_argument("C:\\Some Path\\With Spaces")
    assert encoded == '^"C:\\Some Path\\With Spaces^"'


def test_broadcast_setting_change():
    """
    Test to rehash the Environment variables
    """
    assert win_functions.broadcast_setting_change()


def test_get_user_groups():
    groups = ["Administrators", "Users"]
    with patch("win32net.NetUserGetLocalGroups", return_value=groups):
        assert win_functions.get_user_groups("Administrator") == groups


def test_get_user_groups_sid():
    groups = ["Administrators", "Users"]
    expected = ["S-1-5-32-544", "S-1-5-32-545"]
    with patch("win32net.NetUserGetLocalGroups", return_value=groups):
        assert win_functions.get_user_groups("Administrator", sid=True) == expected


def test_get_user_groups_system():
    groups = ["SYSTEM"]
    with patch("win32net.NetUserGetLocalGroups", return_value=groups):
        assert win_functions.get_user_groups("SYSTEM") == groups


@pytest.mark.skipif(not HAS_WIN32, reason="Requires Win32 libraries")
def test_get_user_groups_unavailable_dc():
    groups = ["Administrators", "Users"]
    win_error = WinError()
    win_error.winerror = 1722
    effect = [win_error, groups]
    with patch("win32net.NetUserGetLocalGroups", side_effect=effect):
        assert win_functions.get_user_groups("Administrator") == groups


@pytest.mark.skipif(not HAS_WIN32, reason="Requires Win32 libraries")
def test_get_user_groups_unknown_dc():
    groups = ["Administrators", "Users"]
    win_error = WinError()
    win_error.winerror = 2453
    effect = [win_error, groups]
    with patch("win32net.NetUserGetLocalGroups", side_effect=effect):
        assert win_functions.get_user_groups("Administrator") == groups


@pytest.mark.skipif(not HAS_WIN32, reason="Requires Win32 libraries")
def test_get_user_groups_missing_permission():
    groups = ["Administrators", "Users"]
    win_error = WinError()
    win_error.winerror = 5
    effect = [win_error, groups]
    with patch("win32net.NetUserGetLocalGroups", side_effect=effect):
        assert win_functions.get_user_groups("Administrator") == groups


@pytest.mark.skipif(not HAS_WIN32, reason="Requires Win32 libraries")
def test_get_user_groups_error():
    win_error = WinError()
    win_error.winerror = 1927
    mock_error = MagicMock(side_effect=win_error)
    with patch("win32net.NetUserGetLocalGroups", side_effect=mock_error):
        with pytest.raises(WinError):
            win_functions.get_user_groups("Administrator")


@pytest.mark.skipif(not HAS_PYWIN, reason="Requires pywintypes libraries")
def test_get_user_groups_local_pywin_error():
    win_error = PyWinError()
    win_error.winerror = 1355
    mock_error = MagicMock(side_effect=win_error)
    with patch("win32net.NetUserGetLocalGroups", side_effect=mock_error):
        with pytest.raises(PyWinError):
            win_functions.get_user_groups("Administrator")


@pytest.mark.skipif(not HAS_PYWIN, reason="Requires pywintypes libraries")
def test_get_user_groups_pywin_error():
    win_error = PyWinError()
    win_error.winerror = 1355
    mock_error = MagicMock(side_effect=win_error)
    with patch("win32net.NetUserGetLocalGroups", side_effect=mock_error):
        with patch("win32net.NetUserGetGroups", side_effect=mock_error):
            with pytest.raises(PyWinError):
                win_functions.get_user_groups("Administrator")


@pytest.mark.skipif(not HAS_PYWIN, reason="Requires pywintypes libraries")
def test_get_sam_name_lookup_fails():
    win_error = PyWinError()
    mock_error = MagicMock(side_effect=win_error)
    with patch("win32security.LookupAccountName", side_effect=mock_error):
        expected = "\\".join([platform.node()[:15].upper(), "junk"])
        result = win_functions.get_sam_name("junk")
        assert result == expected


@pytest.mark.skipif(not HAS_PYWIN, reason="Requires pywintypes libraries")
def test_get_sam_name_everyone():
    expected = "Everyone"
    result = win_functions.get_sam_name("Everyone")
    assert result == expected


@pytest.mark.skipif(not HAS_PYWIN, reason="Requires pywintypes libraries")
def test_get_sam_name(test_user):
    expected = "\\".join([platform.node()[:15], test_user])
    result = win_functions.get_sam_name(test_user)
    assert result.lower() == expected.lower()


@pytest.mark.parametrize(
    "string, expected",
    [
        ("foo", ["foo"]),
        ("foo bar", ["foo", "bar"]),
        ("foo bar=baz", ["foo", "bar=baz"]),
        ('foo "bar baz"', ["foo", "bar baz"]),
        ("C:\\Temp\\test.txt", ["C:\\Temp\\test.txt"]),
        ('"C:\\Temp\\Space Dir\\test.txt"', ["C:\\Temp\\Space Dir\\test.txt"]),
    ],
)
def test_shlex_split(string, expected):
    result = win_functions.shlex_split(string=string)
    assert result == expected



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import os

import pytest

import salt.states.ssh_known_hosts as ssh_known_hosts
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {ssh_known_hosts: {}}


def test_present():
    """
    Test to verify that the specified host is known by the specified user.
    """
    config = ".ssh/known_hosts"
    name = "github.com"
    user = "root"
    enc = "ssh-ed25519"
    key = "AAAAC3NzaC1lZDI1NTE5AAAAIOMqqnkVzrm0SdG6UOoqKLsabgH5C9okWi0dh2l9GKJl"
    fingerprint = "65:96:2d:fc:e8:d5:a9:11:64:0c:0f:ea:00:6e:5b:bd"

    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    with patch.dict(ssh_known_hosts.__opts__, {"test": True}):
        with patch.object(os.path, "isabs", MagicMock(return_value=False)):
            comt = 'If not specifying a "user", specify an absolute "config".'
            ret.update({"comment": comt})
            assert ssh_known_hosts.present(name) == ret

        comt = 'Specify either "key" or "fingerprint", not both.'
        ret.update({"comment": comt})
        assert ssh_known_hosts.present(name, user, key=key, fingerprint=[key]) == ret

        comt = 'Required argument "enc" if using "key" argument.'
        ret.update({"comment": comt})
        assert ssh_known_hosts.present(name, user, key=key) == ret

        mock = MagicMock(side_effect=["exists", "add", "update"])
        with patch.dict(ssh_known_hosts.__salt__, {"ssh.check_known_host": mock}):
            comt = f"Host github.com is already in {config}"
            ret.update({"comment": comt, "result": True})
            assert ssh_known_hosts.present(name, user) == ret

            comt = f"Key for github.com is set to be added to {config}"
            ret.update({"comment": comt, "result": None})
            assert ssh_known_hosts.present(name, user) == ret

            comt = f"Key for github.com is set to be updated in {config}"
            ret.update({"comment": comt})
            assert ssh_known_hosts.present(name, user) == ret

    with patch.dict(ssh_known_hosts.__opts__, {"test": False}):
        result = {"status": "exists", "error": ""}
        mock = MagicMock(return_value=result)
        with patch.dict(ssh_known_hosts.__salt__, {"ssh.set_known_host": mock}):
            comt = f"github.com already exists in {config}"
            ret.update({"comment": comt, "result": True})
            assert ssh_known_hosts.present(name, user) == ret

        result = {"status": "error", "error": ""}
        mock = MagicMock(return_value=result)
        with patch.dict(ssh_known_hosts.__salt__, {"ssh.set_known_host": mock}):
            ret.update({"comment": "", "result": False})
            assert ssh_known_hosts.present(name, user) == ret

        result = {
            "status": "updated",
            "error": "",
            "new": [{"fingerprint": fingerprint, "key": key}],
            "old": "",
        }
        mock = MagicMock(return_value=result)
        with patch.dict(ssh_known_hosts.__salt__, {"ssh.set_known_host": mock}):
            comt = f"{name}'s key saved to {config} (key: {key})"
            ret.update(
                {
                    "comment": comt,
                    "result": True,
                    "changes": {
                        "new": [{"fingerprint": fingerprint, "key": key}],
                        "old": "",
                    },
                }
            )
            assert ssh_known_hosts.present(name, user, key=key) == ret

            comt = f"{name}'s key saved to {config} (fingerprint: {fingerprint})"
            ret.update({"comment": comt})
            assert ssh_known_hosts.present(name, user, fingerprint=fingerprint) == ret

            comt = f"{name}'s {enc} key saved to {config} (key: {key})"
            ret.update({"comment": comt})
            assert ssh_known_hosts.present(name, user, enc=enc) == ret

            comt = f"{name}'s key(s) saved to {config}"
            ret.update({"comment": comt})
            assert ssh_known_hosts.present(name, user) == ret


def test_absent():
    """
    Test to verifies that the specified host is not known by the given user.
    """
    name = "github.com"
    user = "root"

    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    with patch.object(os.path, "isabs", MagicMock(return_value=False)):
        comt = 'If not specifying a "user", specify an absolute "config".'
        ret.update({"comment": comt})
        assert ssh_known_hosts.absent(name) == ret

    mock = MagicMock(return_value=False)
    with patch.dict(ssh_known_hosts.__salt__, {"ssh.get_known_host_entries": mock}):
        comt = "Host is already absent"
        ret.update({"comment": comt, "result": True})
        assert ssh_known_hosts.absent(name, user) == ret

    mock = MagicMock(return_value=True)
    with patch.dict(ssh_known_hosts.__salt__, {"ssh.get_known_host_entries": mock}):
        with patch.dict(ssh_known_hosts.__opts__, {"test": True}):
            comt = "Key for github.com is set to be removed from .ssh/known_hosts"
            ret.update({"comment": comt, "result": None})
            assert ssh_known_hosts.absent(name, user) == ret

        with patch.dict(ssh_known_hosts.__opts__, {"test": False}):
            result = {"status": "error", "error": ""}
            mock = MagicMock(return_value=result)
            with patch.dict(ssh_known_hosts.__salt__, {"ssh.rm_known_host": mock}):
                ret.update({"comment": "", "result": False})
                assert ssh_known_hosts.absent(name, user) == ret

            result = {"status": "removed", "error": "", "comment": "removed"}
            mock = MagicMock(return_value=result)
            with patch.dict(ssh_known_hosts.__salt__, {"ssh.rm_known_host": mock}):
                ret.update(
                    {
                        "comment": "removed",
                        "result": True,
                        "changes": {"new": None, "old": True},
                    }
                )
                assert ssh_known_hosts.absent(name, user) == ret



"""
Manage groups on Linux, OpenBSD and NetBSD

.. important::
    If you feel that Salt should be using this module to manage groups on a
    minion, and it is using a different module (or gives an error similar to
    *'group.info' is not available*), see :ref:`here
    <module-provider-override>`.
"""

import functools
import logging
import os

import salt.utils.files
import salt.utils.path
import salt.utils.stringutils
from salt.exceptions import CommandExecutionError

try:
    import grp
except ImportError:
    pass

log = logging.getLogger(__name__)


# Define the module's virtual name
__virtualname__ = "group"


def __virtual__():
    """
    Set the user module if the kernel is Linux or OpenBSD
    """
    if __grains__["kernel"] in ("Linux", "OpenBSD", "NetBSD"):
        return __virtualname__
    return (
        False,
        "The groupadd execution module cannot be loaded: "
        " only available on Linux, OpenBSD and NetBSD",
    )


def _which(cmd):
    """
    Utility function wrapper to error out early if a command is not found
    """
    _cmd = salt.utils.path.which(cmd)
    if not _cmd:
        raise CommandExecutionError(f"Command '{cmd}' cannot be found")
    return _cmd


def add(name, gid=None, system=False, root=None, non_unique=False):
    """
    .. versionchanged:: 3006.0

    Add the specified group

    name
        Name of the new group

    gid
        Use GID for the new group

    system
        Create a system account

    root
        Directory to chroot into

    non_unique
        Allow creating groups with duplicate (non-unique) GIDs

        .. versionadded:: 3006.0

    CLI Example:

    .. code-block:: bash

        salt '*' group.add foo 3456
    """
    cmd = [_which("groupadd")]
    if gid:
        cmd.append(f"-g {gid}")
        if non_unique:
            cmd.append("-o")
    if system and __grains__["kernel"] != "OpenBSD":
        cmd.append("-r")

    if root is not None:
        cmd.extend(("-R", root))

    cmd.append(name)

    ret = __salt__["cmd.run_all"](cmd, python_shell=False)

    return not ret["retcode"]


def delete(name, root=None):
    """
    Remove the named group

    name
        Name group to delete

    root
        Directory to chroot into

    CLI Example:

    .. code-block:: bash

        salt '*' group.delete foo
    """
    cmd = [_which("groupdel")]

    if root is not None:
        cmd.extend(("-R", root))

    cmd.append(name)

    ret = __salt__["cmd.run_all"](cmd, python_shell=False)

    return not ret["retcode"]


def info(name, root=None):
    """
    Return information about a group

    name
        Name of the group

    root
        Directory to chroot into

    CLI Example:

    .. code-block:: bash

        salt '*' group.info foo
    """
    if root is not None:
        getgrnam = functools.partial(_getgrnam, root=root)
    else:
        getgrnam = functools.partial(grp.getgrnam)

    try:
        grinfo = getgrnam(name)
    except KeyError:
        return {}
    else:
        return _format_info(grinfo)


def _format_info(data):
    """
    Return formatted information in a pretty way.
    """
    return {
        "name": data.gr_name,
        "passwd": data.gr_passwd,
        "gid": data.gr_gid,
        "members": data.gr_mem,
    }


def getent(refresh=False, root=None):
    """
    Return info on all groups

    refresh
        Force a refresh of group information

    root
        Directory to chroot into

    CLI Example:

    .. code-block:: bash

        salt '*' group.getent
    """
    if "group.getent" in __context__ and not refresh:
        return __context__["group.getent"]

    ret = []
    if root is not None:
        getgrall = functools.partial(_getgrall, root=root)
    else:
        getgrall = functools.partial(grp.getgrall)

    for grinfo in getgrall():
        ret.append(_format_info(grinfo))
    __context__["group.getent"] = ret
    return ret


def _chattrib(name, key, value, param, root=None):
    """
    Change an attribute for a named user
    """
    pre_info = info(name, root=root)
    if not pre_info:
        return False

    if value == pre_info[key]:
        return True

    cmd = [_which("groupmod")]

    if root is not None:
        cmd.extend(("-R", root))

    cmd.extend((param, value, name))

    __salt__["cmd.run"](cmd, python_shell=False)
    return info(name, root=root).get(key) == value


def chgid(name, gid, root=None, non_unique=False):
    """
    .. versionchanged:: 3006.0

    Change the gid for a named group

    name
        Name of the group to modify

    gid
        Change the group ID to GID

    root
        Directory to chroot into

    non_unique
        Allow modifying groups with duplicate (non-unique) GIDs

        .. versionadded:: 3006.0

    CLI Example:

    .. code-block:: bash

        salt '*' group.chgid foo 4376
    """
    param = "-g"
    if non_unique:
        param = "-og"
    return _chattrib(name, "gid", gid, param, root=root)


def adduser(name, username, root=None):
    """
    Add a user in the group.

    name
        Name of the group to modify

    username
        Username to add to the group

    root
        Directory to chroot into

    CLI Example:

    .. code-block:: bash

         salt '*' group.adduser foo bar

    Verifies if a valid username 'bar' as a member of an existing group 'foo',
    if not then adds it.
    """
    on_redhat_5 = (
        __grains__.get("os_family") == "RedHat"
        and __grains__.get("osmajorrelease") == "5"
    )
    on_suse_11 = (
        __grains__.get("os_family") == "Suse"
        and __grains__.get("osmajorrelease") == "11"
    )

    if __grains__["kernel"] == "Linux":
        if on_redhat_5:
            cmd = [_which("gpasswd"), "-a", username, name]
        elif on_suse_11:
            cmd = [_which("usermod"), "-A", name, username]
        else:
            cmd = [_which("gpasswd"), "--add", username, name]
        if root is not None:
            cmd.extend(("--root", root))
    else:
        cmd = [_which("usermod"), "-G", name, username]
        if root is not None:
            cmd.extend(("-R", root))

    retcode = __salt__["cmd.retcode"](cmd, python_shell=False)

    return not retcode


def deluser(name, username, root=None):
    """
    Remove a user from the group.

    name
        Name of the group to modify

    username
        Username to delete from the group

    root
        Directory to chroot into

    CLI Example:

    .. code-block:: bash

         salt '*' group.deluser foo bar

    Removes a member user 'bar' from a group 'foo'. If group is not present
    then returns True.
    """
    on_redhat_5 = (
        __grains__.get("os_family") == "RedHat"
        and __grains__.get("osmajorrelease") == "5"
    )
    on_suse_11 = (
        __grains__.get("os_family") == "Suse"
        and __grains__.get("osmajorrelease") == "11"
    )

    grp_info = __salt__["group.info"](name)
    try:
        if username in grp_info["members"]:
            if __grains__["kernel"] == "Linux":
                if on_redhat_5:
                    cmd = [_which("gpasswd"), "-d", username, name]
                elif on_suse_11:
                    cmd = [_which("usermod"), "-R", name, username]
                else:
                    cmd = [_which("gpasswd"), "--del", username, name]
                if root is not None:
                    cmd.extend(("--root", root))
                retcode = __salt__["cmd.retcode"](cmd, python_shell=False)
            elif __grains__["kernel"] == "OpenBSD":
                out = __salt__["cmd.run_stdout"](
                    f"id -Gn {username}", python_shell=False
                )
                cmd = [_which("usermod"), "-S"]
                cmd.append(",".join([g for g in out.split() if g != str(name)]))
                cmd.append(f"{username}")
                retcode = __salt__["cmd.retcode"](cmd, python_shell=False)
            else:
                log.error("group.deluser is not yet supported on this platform")
                return False
            return not retcode
        else:
            return True
    except CommandExecutionError:
        raise
    except Exception:  # pylint: disable=broad-except
        return True


def members(name, members_list, root=None):
    """
    Replaces members of the group with a provided list.

    name
        Name of the group to modify

    members_list
        Username list to set into the group

    root
        Directory to chroot into

    CLI Example:

    .. code-block:: bash

        salt '*' group.members foo 'user1,user2,user3,...'

    Replaces a membership list for a local group 'foo'.
        foo:x:1234:user1,user2,user3,...
    """
    on_redhat_5 = (
        __grains__.get("os_family") == "RedHat"
        and __grains__.get("osmajorrelease") == "5"
    )
    on_suse_11 = (
        __grains__.get("os_family") == "Suse"
        and __grains__.get("osmajorrelease") == "11"
    )

    if __grains__["kernel"] == "Linux":
        if on_redhat_5:
            cmd = [_which("gpasswd"), "-M", members_list, name]
        elif on_suse_11:
            for old_member in __salt__["group.info"](name).get("members"):
                __salt__["cmd.run"](
                    "{} -R {} {}".format(_which("groupmod"), old_member, name),
                    python_shell=False,
                )
            cmd = [_which("groupmod"), "-A", members_list, name]
        else:
            cmd = [_which("gpasswd"), "--members", members_list, name]
        if root is not None:
            cmd.extend(("--root", root))
        retcode = __salt__["cmd.retcode"](cmd, python_shell=False)
    elif __grains__["kernel"] == "OpenBSD":
        retcode = 1
        grp_info = __salt__["group.info"](name)
        if grp_info and name in grp_info["name"]:
            __salt__["cmd.run"](
                "{} {}".format(_which("groupdel"), name), python_shell=False
            )
            __salt__["cmd.run"](
                "{} -g {} {}".format(_which("groupadd"), grp_info["gid"], name),
                python_shell=False,
            )
            for user in members_list.split(","):
                if user:
                    retcode = __salt__["cmd.retcode"](
                        [_which("usermod"), "-G", name, user], python_shell=False
                    )
                    if not retcode == 0:
                        break
                # provided list is '': users previously deleted from group
                else:
                    retcode = 0
    else:
        log.error("group.members is not yet supported on this platform")
        return False

    return not retcode


def _getgrnam(name, root=None):
    """
    Alternative implementation for getgrnam, that use only /etc/group
    """
    root = root or "/"
    passwd = os.path.join(root, "etc/group")
    with salt.utils.files.fopen(passwd) as fp_:
        for line in fp_:
            line = salt.utils.stringutils.to_unicode(line)
            comps = line.strip().split(":")
            if len(comps) < 4:
                log.debug("Ignoring group line: %s", line)
                continue
            if comps[0] == name:
                # Generate a getpwnam compatible output
                comps[2] = int(comps[2])
                comps[3] = comps[3].split(",") if comps[3] else []
                return grp.struct_group(comps)
    raise KeyError(f"getgrnam(): name not found: {name}")


def _getgrall(root=None):
    """
    Alternative implemetantion for getgrall, that use only /etc/group
    """
    root = root or "/"
    passwd = os.path.join(root, "etc/group")
    with salt.utils.files.fopen(passwd) as fp_:
        for line in fp_:
            line = salt.utils.stringutils.to_unicode(line)
            comps = line.strip().split(":")
            if len(comps) < 4:
                log.debug("Ignoring group line: %s", line)
                continue
            # Generate a getgrall compatible output
            comps[2] = int(comps[2])
            comps[3] = comps[3].split(",") if comps[3] else []
            yield grp.struct_group(comps)



import logging
import subprocess

import pytest
from pytestskipmarkers.utils import platform
from saltfactories.utils.functional import MultiStateResult

pytestmark = [
    pytest.mark.skip_on_windows,
]

log = logging.getLogger(__name__)


CHECK_IMPORTS_SLS_CONTENTS = """
#!py
import importlib
import sys

def run():
    config = {}
    for module in [
        '_logging',
        'acl',
        'auth',
        'beacons',
        'cache',
        'cli',
        'client',
        'cloud',
        'config',
        'daemons',
        'defaults',
        'engines',
        'executors',
        'ext',
        'fileserver',
        'grains',
        'matchers',
        'metaproxy',
        'modules',
        'netapi',
        'output',
        'pillar',
        'platform',
        'proxy',
        'queues',
        'renderers',
        'returners',
        'roster',
        'runners',
        'sdb',
        'serializers',
        'spm',
        'states',
        'templates',
        'thorium',
        'tokens',
        'tops',
        'transport',
        'utils',
        'wheel',
    ]:
        import_name = "salt.{}".format(module)
        try:
            importlib.import_module(import_name)
            config[import_name] = {
                'test.succeed_without_changes': [
                    {
                        "name": import_name,
                        'comment': "The '{}' import succeeded.".format(import_name)
                    }
                ]
            }
        except ModuleNotFoundError as err:
            config[import_name] = {
                'test.fail_without_changes': [
                    {
                        "name": import_name,
                        'comment': "The '{}' import failed. The error was: {}".format(import_name, err)
                    }
                ]
            }

    # Import required for all OS'es
    for import_name in [
        "jinja2",
        "telnetlib",
        "yaml",
    ]:
        try:
            importlib.import_module(import_name)
            config[import_name] = {
                'test.succeed_without_changes': [
                    {
                        "name": import_name,
                        'comment': "The '{}' import succeeded.".format(import_name)
                    }
                ]
            }
        except ModuleNotFoundError as err:
            config[import_name] = {
                'test.fail_without_changes': [
                    {
                        "name": import_name,
                        'comment': "The '{}' import failed. The error was: {}".format(import_name, err)
                    }
                ]
            }

    # Windows specific requirements (I think, there may be some for other OSes in here)
    if sys.platform == "win32":
        for import_name in [
            "cffi",
            "clr_loader",
            "lxml",
            "pythonnet",
            "pytz",
            "pywintypes",
            "timelib",
            "win32",
            "wmi",
            "xmltodict",
        ]:
            try:
                importlib.import_module(import_name)
                config[import_name] = {
                    'test.succeed_without_changes': [
                        {
                            "name": import_name,
                            'comment': "The '{}' import succeeded.".format(import_name)
                        }
                    ]
                }
            except ModuleNotFoundError as err:
                config[import_name] = {
                    'test.fail_without_changes': [
                        {
                            "name": import_name,
                            'comment': "The '{}' import failed. The error was: {}".format(import_name, err)
                        }
                    ]
                }
    return config
"""


@pytest.fixture
def state_name(salt_master):
    name = "check-imports"
    with salt_master.state_tree.base.temp_file(
        f"{name}.sls", CHECK_IMPORTS_SLS_CONTENTS
    ):
        if not platform.is_windows() and not platform.is_darwin():
            subprocess.run(
                [
                    "chown",
                    "-R",
                    "salt:salt",
                    str(salt_master.state_tree.base.write_path),
                ],
                check=False,
            )
        yield name


def test_check_imports(salt_cli, salt_minion, state_name):
    """
    Test imports
    """
    ret = salt_cli.run("state.sls", state_name, minion_tgt=salt_minion.id)
    assert ret.returncode == 0
    assert ret.data
    result = MultiStateResult(raw=ret.data)
    for state_ret in result:
        assert state_ret.result is True



"""
Test the win_wusa execution module
"""

import pytest

import salt.modules.win_wusa as win_wusa
from salt.exceptions import CommandExecutionError
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


@pytest.mark.skip_unless_on_windows
class WinWusaTestCase(TestCase, LoaderModuleMockMixin):
    """
    test the functions in the win_wusa execution module
    """

    def setup_loader_modules(self):
        return {win_wusa: {}}

    def test_is_installed_false(self):
        """
        test is_installed function when the KB is not installed
        """
        mock_retcode = MagicMock(return_value=1)
        with patch.dict(win_wusa.__salt__, {"cmd.retcode": mock_retcode}):
            self.assertFalse(win_wusa.is_installed("KB123456"))

    def test_is_installed_true(self):
        """
        test is_installed function when the KB is installed
        """
        mock_retcode = MagicMock(return_value=0)
        with patch.dict(win_wusa.__salt__, {"cmd.retcode": mock_retcode}):
            self.assertTrue(win_wusa.is_installed("KB123456"))

    def test_list(self):
        """
        test list function
        """
        ret = {
            "pid": 1,
            "retcode": 0,
            "stderr": "",
            "stdout": '[{"HotFixID": "KB123456"}, {"HotFixID": "KB123457"}]',
        }
        mock_all = MagicMock(return_value=ret)
        with patch.dict(win_wusa.__salt__, {"cmd.run_all": mock_all}):
            expected = ["KB123456", "KB123457"]
            returned = win_wusa.list()
            self.assertListEqual(expected, returned)

    def test_install(self):
        """
        test install function
        """
        mock_retcode = MagicMock(return_value=0)
        path = "C:\\KB123456.msu"
        with patch.dict(win_wusa.__salt__, {"cmd.retcode": mock_retcode}):
            self.assertTrue(win_wusa.install(path))
        mock_retcode.assert_called_once_with(
            ["wusa.exe", path, "/quiet", "/norestart"], ignore_retcode=True
        )

    def test_install_restart(self):
        """
        test install function with restart=True
        """
        mock_retcode = MagicMock(return_value=0)
        path = "C:\\KB123456.msu"
        with patch.dict(win_wusa.__salt__, {"cmd.retcode": mock_retcode}):
            self.assertTrue(win_wusa.install(path, restart=True))
        mock_retcode.assert_called_once_with(
            ["wusa.exe", path, "/quiet", "/forcerestart"], ignore_retcode=True
        )

    def test_install_already_installed(self):
        """
        test install function when KB already installed
        """
        retcode = 2359302
        mock_retcode = MagicMock(return_value=retcode)
        path = "C:\\KB123456.msu"
        name = "KB123456.msu"
        with patch.dict(win_wusa.__salt__, {"cmd.retcode": mock_retcode}):
            with self.assertRaises(CommandExecutionError) as excinfo:
                win_wusa.install(path)
        mock_retcode.assert_called_once_with(
            ["wusa.exe", path, "/quiet", "/norestart"], ignore_retcode=True
        )
        self.assertEqual(
            "{} is already installed. Additional info follows:\n\n{}".format(
                name, retcode
            ),
            excinfo.exception.strerror,
        )

    def test_install_reboot_needed(self):
        """
        test install function when KB need a reboot
        """
        retcode = 3010
        mock_retcode = MagicMock(return_value=retcode)
        path = "C:\\KB123456.msu"
        name = "KB123456.msu"
        with patch.dict(win_wusa.__salt__, {"cmd.retcode": mock_retcode}):
            with self.assertRaises(CommandExecutionError) as excinfo:
                win_wusa.install(path)
        mock_retcode.assert_called_once_with(
            ["wusa.exe", path, "/quiet", "/norestart"], ignore_retcode=True
        )
        self.assertEqual(
            "{} correctly installed but server reboot is needed to complete"
            " installation. Additional info follows:\n\n{}".format(name, retcode),
            excinfo.exception.strerror,
        )

    def test_install_error_87(self):
        """
        test install function when error 87 returned
        """
        retcode = 87
        mock_retcode = MagicMock(return_value=retcode)
        path = "C:\\KB123456.msu"
        with patch.dict(win_wusa.__salt__, {"cmd.retcode": mock_retcode}):
            with self.assertRaises(CommandExecutionError) as excinfo:
                win_wusa.install(path)
        mock_retcode.assert_called_once_with(
            ["wusa.exe", path, "/quiet", "/norestart"], ignore_retcode=True
        )
        self.assertEqual(
            f"Unknown error. Additional info follows:\n\n{retcode}",
            excinfo.exception.strerror,
        )

    def test_install_error_other(self):
        """
        test install function on other unknown error
        """
        mock_retcode = MagicMock(return_value=1234)
        path = "C:\\KB123456.msu"
        with patch.dict(win_wusa.__salt__, {"cmd.retcode": mock_retcode}):
            with self.assertRaises(CommandExecutionError) as excinfo:
                win_wusa.install(path)
        mock_retcode.assert_called_once_with(
            ["wusa.exe", path, "/quiet", "/norestart"], ignore_retcode=True
        )
        self.assertEqual("Unknown error: 1234", excinfo.exception.strerror)

    def test_uninstall_kb(self):
        """
        test uninstall function passing kb name
        """
        mock_retcode = MagicMock(return_value=0)
        kb = "KB123456"
        with patch.dict(win_wusa.__salt__, {"cmd.retcode": mock_retcode}), patch(
            "os.path.exists", MagicMock(return_value=False)
        ):
            self.assertTrue(win_wusa.uninstall(kb))
        mock_retcode.assert_called_once_with(
            [
                "wusa.exe",
                "/uninstall",
                "/quiet",
                f"/kb:{kb[2:]}",
                "/norestart",
            ],
            ignore_retcode=True,
        )

    def test_uninstall_path(self):
        """
        test uninstall function passing full path to .msu file
        """
        mock_retcode = MagicMock(return_value=0)
        path = "C:\\KB123456.msu"
        with patch.dict(win_wusa.__salt__, {"cmd.retcode": mock_retcode}), patch(
            "os.path.exists", MagicMock(return_value=True)
        ):
            self.assertTrue(win_wusa.uninstall(path))
        mock_retcode.assert_called_once_with(
            ["wusa.exe", "/uninstall", "/quiet", path, "/norestart"],
            ignore_retcode=True,
        )

    def test_uninstall_path_restart(self):
        """
        test uninstall function with full path and restart=True
        """
        mock_retcode = MagicMock(return_value=0)
        path = "C:\\KB123456.msu"
        with patch.dict(win_wusa.__salt__, {"cmd.retcode": mock_retcode}), patch(
            "os.path.exists", MagicMock(return_value=True)
        ):
            self.assertTrue(win_wusa.uninstall(path, restart=True))
        mock_retcode.assert_called_once_with(
            ["wusa.exe", "/uninstall", "/quiet", path, "/forcerestart"],
            ignore_retcode=True,
        )

    def test_uninstall_already_uninstalled(self):
        """
        test uninstall function when KB already uninstalled
        """
        retcode = 2359303
        mock_retcode = MagicMock(return_value=retcode)
        kb = "KB123456"
        with patch.dict(win_wusa.__salt__, {"cmd.retcode": mock_retcode}):
            with self.assertRaises(CommandExecutionError) as excinfo:
                win_wusa.uninstall(kb)
        mock_retcode.assert_called_once_with(
            [
                "wusa.exe",
                "/uninstall",
                "/quiet",
                f"/kb:{kb[2:]}",
                "/norestart",
            ],
            ignore_retcode=True,
        )
        self.assertEqual(
            f"{kb} not installed. Additional info follows:\n\n{retcode}",
            excinfo.exception.strerror,
        )

    def test_uninstall_path_error_other(self):
        """
        test uninstall function with unknown error
        """
        mock_retcode = MagicMock(return_value=1234)
        path = "C:\\KB123456.msu"
        with patch.dict(win_wusa.__salt__, {"cmd.retcode": mock_retcode}), patch(
            "os.path.exists", MagicMock(return_value=True)
        ), self.assertRaises(CommandExecutionError) as excinfo:
            win_wusa.uninstall(path)
        mock_retcode.assert_called_once_with(
            ["wusa.exe", "/uninstall", "/quiet", path, "/norestart"],
            ignore_retcode=True,
        )
        self.assertEqual("Unknown error: 1234", excinfo.exception.strerror)



"""
Test Salt Pkg Downloads
"""

import contextlib
import logging
import os
import pathlib
import shutil
import time

import packaging.version
import psutil
import pytest
from pytestskipmarkers.utils import platform

log = logging.getLogger(__name__)


def get_salt_test_commands():
    salt_release = get_salt_release()
    if platform.is_windows():
        if packaging.version.parse(salt_release) > packaging.version.parse("3005"):
            salt_test_commands = [
                ["salt-call.exe", "--local", "test.versions"],
                ["salt-call.exe", "--local", "grains.items"],
                ["salt-minion.exe", "--version"],
            ]
        else:
            salt_test_commands = [
                ["salt-call.bat", "--local", "test.versions"],
                ["salt-call.bat", "--local", "grains.items"],
                ["salt.bat", "--version"],
                ["salt-master.bat", "--version"],
                ["salt-minion.bat", "--version"],
                ["salt-ssh.bat", "--version"],
                ["salt-syndic.bat", "--version"],
                ["salt-api.bat", "--version"],
                ["salt-cloud.bat", "--version"],
            ]
    else:
        salt_test_commands = [
            ["salt-call", "--local", "test.versions"],
            ["salt-call", "--local", "grains.items"],
            ["salt", "--version"],
            ["salt-master", "--version"],
            ["salt-minion", "--version"],
            ["salt-ssh", "--version"],
            ["salt-syndic", "--version"],
            ["salt-api", "--version"],
            ["salt-cloud", "--version"],
        ]
    return salt_test_commands


@pytest.fixture(scope="module")
def root_url(salt_release):
    default_url = "https://packages.broadcom.com/artifactory"
    repo_domain = os.environ.get("SALT_REPO_DOMAIN_RELEASE", default_url)
    log.info("Repository Root URL: %s", repo_domain)
    return repo_domain


@pytest.fixture(scope="module")
def package_type():
    return os.environ.get("DOWNLOAD_TEST_PACKAGE_TYPE")


def get_salt_release():
    salt_release = os.environ.get("SALT_RELEASE")
    pkg_test_type = os.environ.get("PKG_TEST_TYPE", "install")
    if salt_release is None:
        if pkg_test_type == "download-pkgs":
            log.warning(
                "Setting salt release to 3006.0 which is probably not what you want."
            )
        salt_release = "3006.0"
    if pkg_test_type == "download-pkgs":
        if packaging.version.parse(salt_release) < packaging.version.parse("3006.0"):
            log.warning("The salt release being tested, %r looks off.", salt_release)
    return salt_release


@pytest.fixture(scope="module")
def salt_release():
    yield get_salt_release()


@pytest.fixture(scope="module")
def onedir_install_path(tmp_path_factory):
    install_path = tmp_path_factory.mktemp("onedir_install")
    yield install_path
    shutil.rmtree(install_path, ignore_errors=True)


@pytest.fixture(scope="module")
def _setup_system(
    grains,
    shell,
    root_url,
    salt_release,
    package_type,
    tmp_path_factory,
    onedir_install_path,
):
    downloads_path = tmp_path_factory.mktemp("downloads")
    try:
        # Windows is a special case, because sometimes we need to uninstall the packages
        if grains["os_family"] == "Windows":
            with setup_windows(
                shell,
                root_url=root_url,
                salt_release=salt_release,
                downloads_path=downloads_path,
                package_type=package_type,
                onedir_install_path=onedir_install_path,
            ):
                yield
        else:
            if grains["os_family"] == "MacOS":
                setup_macos(
                    shell,
                    root_url=root_url,
                    salt_release=salt_release,
                    downloads_path=downloads_path,
                    package_type=package_type,
                    onedir_install_path=onedir_install_path,
                )
            elif grains["os"] == "Amazon":
                setup_redhat_family(
                    shell,
                    os_name=grains["os"].lower(),
                    os_version=grains["osmajorrelease"],
                    root_url=root_url,
                    salt_release=salt_release,
                    downloads_path=downloads_path,
                )
            elif grains["os"] == "Fedora":
                setup_redhat_family(
                    shell,
                    os_name=grains["os"].lower(),
                    os_version=grains["osmajorrelease"],
                    root_url=root_url,
                    salt_release=salt_release,
                    downloads_path=downloads_path,
                )
            elif grains["os"] == "VMware Photon OS":
                setup_redhat_family(
                    shell,
                    os_name="photon",
                    os_version=grains["osmajorrelease"],
                    root_url=root_url,
                    salt_release=salt_release,
                    downloads_path=downloads_path,
                )
            elif grains["os_family"] == "RedHat":
                setup_redhat_family(
                    shell,
                    os_name="redhat",
                    os_version=grains["osmajorrelease"],
                    root_url=root_url,
                    salt_release=salt_release,
                    downloads_path=downloads_path,
                )
            elif grains["os_family"] == "Debian":
                setup_debian_family(
                    shell,
                    os_codename=grains["oscodename"],
                    root_url=root_url,
                    salt_release=salt_release,
                    downloads_path=downloads_path,
                    package_type=package_type,
                    onedir_install_path=onedir_install_path,
                )
            else:
                pytest.fail("Don't know how to handle %s", grains["osfinger"])
            yield
    finally:
        shutil.rmtree(downloads_path, ignore_errors=True)


def setup_redhat_family(
    shell,
    os_name,
    os_version,
    root_url,
    salt_release,
    downloads_path,
):
    arch = os.environ.get("SALT_REPO_ARCH") or "x86_64"

    if os_name == "photon":
        os_version = f"{os_version}.0"

    repo_url_base = f"{root_url}/saltproject-rpm"

    # Download the salt.repo
    # It contains the gpg key url so we don't need to download it here
    salt_repo_url = "https://github.com/saltstack/salt-install-guide/releases/latest/download/salt.repo"
    repo_file = pytest.helpers.download_file(
        salt_repo_url, downloads_path / "salt.repo"
    )

    commands = [
        ("mv", str(repo_file), "/etc/yum.repos.d/salt.repo"),
        ("yum", "clean", "all" if os_name == "photon" else "expire-cache"),
        (
            "yum",
            "install",
            "-y",
            f"salt-master-{salt_release}",
            f"salt-minion-{salt_release}",
            f"salt-ssh-{salt_release}",
            f"salt-syndic-{salt_release}",
            f"salt-cloud-{salt_release}",
            f"salt-api-{salt_release}",
            f"salt-debuginfo-{salt_release}",
        ),
    ]

    for cmd in commands:
        ret = shell.run(*cmd, check=False)
        if ret.returncode != 0:
            pytest.fail(f"Failed to run '{' '.join(cmd)!r}':\n{ret}")


def setup_debian_family(
    shell,
    os_codename,
    root_url,
    salt_release,
    downloads_path,
    package_type,
    onedir_install_path,
):
    arch = os.environ.get("SALT_REPO_ARCH") or "amd64"
    ret = shell.run("apt-get", "update", "-y", check=False)
    if ret.returncode != 0:
        pytest.fail(str(ret))

    if package_type == "package":
        if arch == "aarch64":
            arch = "arm64"
        elif arch == "x86_64":
            arch = "amd64"

        repo_url_base = f"{root_url}/saltproject-deb/"
        gpg_file_url = "https://packages.broadcom.com/artifactory/api/security/keypair/SaltProjectKey/public"
        gpg_key_name = "SALT-PROJECT-GPG-PUBKEY-2023.pub"

        try:
            pytest.helpers.download_file(gpg_file_url, downloads_path / gpg_key_name)
        except Exception as exc:  # pylint: disable=broad-except
            pytest.fail(f"Failed to download {gpg_file_url}: {exc}")

        salt_sources_path = downloads_path / "salt.list"
        salt_sources_path.write_text(
            f"deb [signed-by=/usr/share/keyrings/{gpg_key_name} arch={arch}] {repo_url_base} {os_codename} main\n"
        )
        commands = [
            (
                "mv",
                str(downloads_path / gpg_key_name),
                f"/usr/share/keyrings/{gpg_key_name}",
            ),
            (
                "mv",
                str(salt_sources_path),
                "/etc/apt/sources.list.d/salt.list",
            ),
            ("apt-get", "install", "-y", "ca-certificates"),
            ("update-ca-certificates",),
            ("apt-get", "update"),
            (
                "apt-get",
                "install",
                "-y",
                "salt-master",
                "salt-minion",
                "salt-ssh",
                "salt-syndic",
                "salt-cloud",
                "salt-api",
                "salt-dbg",
            ),
        ]
        for cmd in commands:
            ret = shell.run(*cmd)
            if ret.returncode != 0:
                pytest.fail(str(ret))
    else:
        # We are testing the onedir download
        onedir_name = f"salt-{salt_release}-onedir-linux-{arch}.tar.xz"
        onedir_url = (
            f"{root_url}/saltproject-generic/onedir/{salt_release}/{onedir_name}"
        )
        onedir_location = downloads_path / onedir_name
        onedir_extracted = onedir_install_path

        try:
            pytest.helpers.download_file(onedir_url, onedir_location)
        except Exception as exc:  # pylint: disable=broad-except
            pytest.fail(f"Failed to download {onedir_url}: {exc}")

        shell.run("tar", "xvf", str(onedir_location), "-C", str(onedir_extracted))


def setup_macos(
    shell,
    root_url,
    salt_release,
    downloads_path,
    package_type,
    onedir_install_path,
):
    arch = os.environ.get("SALT_REPO_ARCH") or "x86_64"
    if package_type == "package":
        mac_pkg = f"salt-{salt_release}-py3-{arch}.pkg"
        mac_pkg_url = f"{root_url}/saltproject-generic/macos/{salt_release}/{mac_pkg}"

        mac_pkg_path = downloads_path / mac_pkg
        pytest.helpers.download_file(mac_pkg_url, mac_pkg_path)

        ret = shell.run(
            "installer",
            "-pkg",
            str(mac_pkg_path),
            "-target",
            "/",
            check=False,
        )
        assert ret.returncode == 0, ret
    else:
        # We are testing the onedir download
        onedir_name = f"salt-{salt_release}-onedir-macos-{arch}.tar.xz"
        onedir_url = f"{root_url}/onedir/{salt_release}/{onedir_name}"
        onedir_location = downloads_path / onedir_name
        onedir_extracted = onedir_install_path

        try:
            pytest.helpers.download_file(onedir_url, onedir_location)
        except Exception as exc:  # pylint: disable=broad-except
            pytest.fail(f"Failed to download {onedir_url}: {exc}")

        shell.run("tar", "xvf", str(onedir_location), "-C", str(onedir_extracted))


@contextlib.contextmanager
def setup_windows(
    shell,
    root_url,
    salt_release,
    downloads_path,
    package_type,
    onedir_install_path,
    timeout=300,
):
    try:
        arch = os.environ.get("SALT_REPO_ARCH") or "amd64"
        if package_type != "onedir":
            root_dir = pathlib.Path(os.getenv("ProgramFiles"), "Salt Project", "Salt")

            if package_type.lower() == "nsis":
                if arch.lower() != "x86":
                    arch = arch.upper()
                win_pkg = f"Salt-Minion-{salt_release}-Py3-{arch}-Setup.exe"
            else:
                if arch.lower() != "x86":
                    arch = arch.upper()
                win_pkg = f"Salt-Minion-{salt_release}-Py3-{arch}.msi"
            win_pkg_url = (
                f"{root_url}/saltproject-generic/windows/{salt_release}/{win_pkg}"
            )
            ssm_bin = root_dir / "ssm.exe"

            pkg_path = downloads_path / win_pkg

            pytest.helpers.download_file(win_pkg_url, pkg_path)
            if package_type.lower() == "nsis":
                # We need to make sure there are no installer/uninstaller
                # processes running. Uninst.exe launches a 2nd binary
                # (Un.exe or Un_*.exe) Let's get the name of the process
                processes = [
                    win_pkg,
                    "uninst.exe",
                    "Un.exe",
                    "Un_A.exe",
                    "Un_B.exe",
                    "Un_C.exe",
                    "Un_D.exe",
                    "Un_D.exe",
                    "Un_F.exe",
                    "Un_G.exe",
                ]
                proc_name = ""
                for proc in processes:
                    try:
                        if proc in (p.name() for p in psutil.process_iter()):
                            proc_name = proc
                    except psutil.NoSuchProcess:
                        continue

                # We need to give the process time to exit. We'll timeout after
                # 5 minutes or whatever timeout is set to
                if proc_name:
                    elapsed_time = 0
                    while elapsed_time < timeout:
                        try:
                            if proc_name not in (
                                p.name() for p in psutil.process_iter()
                            ):
                                break
                        except psutil.NoSuchProcess:
                            continue
                        elapsed_time += 0.1
                        time.sleep(0.1)

                # Only run setup when we're sure no other installations are running
                ret = shell.run(str(pkg_path), "/start-minion=0", "/S", check=False)
            else:
                ret = shell.run(
                    "msiexec", "/qn", "/i", str(pkg_path), 'START_MINION=""'
                )
            assert ret.returncode == 0, ret

            log.debug("Removing installed salt-minion service")
            ret = shell.run(
                "cmd",
                "/c",
                str(ssm_bin),
                "remove",
                "salt-minion",
                "confirm",
                check=False,
            )
            assert ret.returncode == 0, ret
        else:
            # We are testing the onedir download
            onedir_name = f"salt-{salt_release}-onedir-windows-{arch}.zip"
            onedir_url = f"{root_url}/onedir/{salt_release}/{onedir_name}"
            onedir_location = downloads_path / onedir_name
            onedir_extracted = onedir_install_path

            try:
                pytest.helpers.download_file(onedir_url, onedir_location)
            except Exception as exc:  # pylint: disable=broad-except
                pytest.fail(f"Failed to download {onedir_url}: {exc}")

            shell.run("unzip", str(onedir_location), "-d", str(onedir_extracted))
        yield
    finally:
        # We need to uninstall the MSI packages, otherwise they will not install correctly
        if package_type.lower() == "msi":
            ret = shell.run("msiexec", "/qn", "/x", str(pkg_path))
            assert ret.returncode == 0, ret


@pytest.fixture(scope="module")
def install_dir(_setup_system, package_type, onedir_install_path):
    if package_type != "onedir":
        if platform.is_windows():
            return pathlib.Path(
                os.getenv("ProgramFiles"), "Salt Project", "Salt"
            ).resolve()
        if platform.is_darwin():
            return pathlib.Path("/opt", "salt")
        return pathlib.Path("/opt", "saltstack", "salt")
    else:
        # We are testing the onedir
        return onedir_install_path / "salt"


@pytest.fixture(scope="module")
def salt_test_command(request, install_dir):
    command = request.param
    command[0] = str(install_dir / command[0])
    return command


@pytest.mark.skip_on_windows(reason="This is flaky on Windows")
@pytest.mark.parametrize("salt_test_command", get_salt_test_commands(), indirect=True)
def test_download(shell, salt_test_command):
    """
    Test downloading of Salt packages and running various commands.
    """
    ret = shell.run(*salt_test_command, check=False)
    assert ret.returncode == 0, ret



"""
Glues the VMware vSphere Execution Module to the VMware ESXi Proxy Minions to the
:mod:`esxi proxymodule <salt.proxy.esxi>`.

.. Warning::
    This module will be deprecated in a future release of Salt. VMware strongly
    recommends using the
    `VMware Salt extensions <https://docs.saltproject.io/salt/extensions/salt-ext-modules-vmware/en/latest/all.html>`_
    instead of the ESXi module. Because the Salt extensions are newer and
    actively supported by VMware, they are more compatible with current versions
    of ESXi and they work well with the latest features in the VMware product
    line.


Depends: :mod:`vSphere Remote Execution Module (salt.modules.vsphere)
<salt.modules.vsphere>`

For documentation on commands that you can direct to an ESXi host via proxy,
look in the documentation for :mod:`salt.modules.vsphere <salt.modules.vsphere>`.

This execution module calls through to a function in the ESXi proxy module
called ``ch_config``, which looks up the function passed in the ``command``
parameter in :mod:`salt.modules.vsphere <salt.modules.vsphere>` and calls it.

To execute commands with an ESXi Proxy Minion using the vSphere Execution Module,
use the ``esxi.cmd <vsphere-function-name>`` syntax. Both args and kwargs needed
for various vsphere execution module functions must be passed through in a kwarg-
type manor.

.. code-block:: bash

    salt 'esxi-proxy' esxi.cmd system_info
    salt 'exsi-proxy' esxi.cmd get_service_policy service_name='ssh'

"""

import logging
from functools import wraps

import salt.utils.platform

log = logging.getLogger(__name__)

__proxyenabled__ = ["esxi"]
__virtualname__ = "esxi"


def __virtual__():
    """
    Only work on proxy
    """
    if salt.utils.platform.is_proxy():
        return __virtualname__
    return (
        False,
        "The esxi execution module failed to load: only available on proxy minions.",
    )


def _deprecation_message(function):
    """
    Decorator wrapper to warn about azurearm deprecation
    """

    @wraps(function)
    def wrapped(*args, **kwargs):
        salt.utils.versions.warn_until(
            "Argon",
            "The 'esxi' functionality in Salt has been deprecated and its "
            "functionality will be removed in version 3008 in favor of the "
            "saltext.vmware Salt Extension. "
            "(https://github.com/saltstack/salt-ext-modules-vmware)",
            category=FutureWarning,
        )
        ret = function(*args, **salt.utils.args.clean_kwargs(**kwargs))
        return ret

    return wrapped


@_deprecation_message
def cmd(command, *args, **kwargs):
    proxy_prefix = __opts__["proxy"]["proxytype"]
    proxy_cmd = proxy_prefix + ".ch_config"

    return __proxy__[proxy_cmd](command, *args, **kwargs)


@_deprecation_message
def get_details():
    return __proxy__["esxi.get_details"]()



"""
A Websockets add-on to saltnado
===============================

.. py:currentmodule:: salt.netapi.rest_tornado.saltnado

:depends:   - tornado Python module

In order to enable saltnado_websockets you must add websockets: True to your
saltnado config block.

.. code-block:: yaml

    rest_tornado:
        # can be any port
        port: 8000
        ssl_crt: /etc/pki/api/certs/server.crt
        # no need to specify ssl_key if cert and key
        # are in one single file
        ssl_key: /etc/pki/api/certs/server.key
        debug: False
        disable_ssl: False
        websockets: True

All Events
----------

Exposes ``all`` "real-time" events from Salt's event bus on a websocket connection.
It should be noted that "Real-time" here means these events are made available
to the server as soon as any salt related action (changes to minions, new jobs etc) happens.
Clients are however assumed to be able to tolerate any network transport related latencies.
Functionality provided by this endpoint is similar to the ``/events`` end point.

The event bus on the Salt master exposes a large variety of things, notably
when executions are started on the master and also when minions ultimately
return their results. This URL provides a real-time window into a running
Salt infrastructure. Uses websocket as the transport mechanism.

Exposes GET method to return websocket connections.
All requests should include an auth token.
A way to obtain obtain authentication tokens is shown below.

.. code-block:: bash

    % curl -si localhost:8000/login \\
        -H "Accept: application/json" \\
        -d username='salt' \\
        -d password='salt' \\
        -d eauth='pam'

Which results in the response

.. code-block:: json

    {
        "return": [{
            "perms": [".*", "@runner", "@wheel"],
            "start": 1400556492.277421,
            "token": "d0ce6c1a37e99dcc0374392f272fe19c0090cca7",
            "expire": 1400599692.277422,
            "user": "salt",
            "eauth": "pam"
        }]
    }

In this example the ``token`` returned is ``d0ce6c1a37e99dcc0374392f272fe19c0090cca7`` and can be included
in subsequent websocket requests (as part of the URL).

The event stream can be easily consumed via JavaScript:

.. code-block:: javascript

    // Note, you must be authenticated!

    // Get the Websocket connection to Salt
    var source = new Websocket('wss://localhost:8000/all_events/d0ce6c1a37e99dcc0374392f272fe19c0090cca7');

    // Get Salt's "real time" event stream.
    source.onopen = function() { source.send('websocket client ready'); };

    // Other handlers
    source.onerror = function(e) { console.debug('error!', e); };

    // e.data represents Salt's "real time" event data as serialized JSON.
    source.onmessage = function(e) { console.debug(e.data); };

    // Terminates websocket connection and Salt's "real time" event stream on the server.
    source.close();

Or via Python, using the Python module
`websocket-client <https://pypi.python.org/pypi/websocket-client/>`_ for example.
Or the tornado
`client <https://tornado.readthedocs.io/en/latest/websocket.html#client-side-support>`_.

.. code-block:: python

    # Note, you must be authenticated!

    from websocket import create_connection

    # Get the Websocket connection to Salt
    ws = create_connection('wss://localhost:8000/all_events/d0ce6c1a37e99dcc0374392f272fe19c0090cca7')

    # Get Salt's "real time" event stream.
    ws.send('websocket client ready')


    # Simple listener to print results of Salt's "real time" event stream.
    # Look at https://pypi.python.org/pypi/websocket-client/ for more examples.
    while listening_to_events:
        print ws.recv()       #  Salt's "real time" event data as serialized JSON.

    # Terminates websocket connection and Salt's "real time" event stream on the server.
    ws.close()

    # Please refer to https://github.com/liris/websocket-client/issues/81 when using a self signed cert

Above examples show how to establish a websocket connection to Salt and activating
real time updates from Salt's event stream by signaling ``websocket client ready``.


Formatted Events
-----------------

Exposes ``formatted`` "real-time" events from Salt's event bus on a websocket connection.
It should be noted that "Real-time" here means these events are made available
to the server as soon as any salt related action (changes to minions, new jobs etc) happens.
Clients are however assumed to be able to tolerate any network transport related latencies.
Functionality provided by this endpoint is similar to the ``/events`` end point.

The event bus on the Salt master exposes a large variety of things, notably
when executions are started on the master and also when minions ultimately
return their results. This URL provides a real-time window into a running
Salt infrastructure. Uses websocket as the transport mechanism.

Formatted events parses the raw "real time" event stream and maintains
a current view of the following:

- minions
- jobs

A change to the minions (such as addition, removal of keys or connection drops)
or jobs is processed and clients are updated.
Since we use salt's presence events to track minions,
please enable ``presence_events``
and set a small value for the ``loop_interval``
in the salt master config file.

Exposes GET method to return websocket connections.
All requests should include an auth token.
A way to obtain obtain authentication tokens is shown below.

.. code-block:: bash

    % curl -si localhost:8000/login \\
        -H "Accept: application/json" \\
        -d username='salt' \\
        -d password='salt' \\
        -d eauth='pam'

Which results in the response

.. code-block:: json

    {
        "return": [{
            "perms": [".*", "@runner", "@wheel"],
            "start": 1400556492.277421,
            "token": "d0ce6c1a37e99dcc0374392f272fe19c0090cca7",
            "expire": 1400599692.277422,
            "user": "salt",
            "eauth": "pam"
        }]
    }

In this example the ``token`` returned is ``d0ce6c1a37e99dcc0374392f272fe19c0090cca7`` and can be included
in subsequent websocket requests (as part of the URL).

The event stream can be easily consumed via JavaScript:

.. code-block:: javascript

    // Note, you must be authenticated!

    // Get the Websocket connection to Salt
    var source = new Websocket('wss://localhost:8000/formatted_events/d0ce6c1a37e99dcc0374392f272fe19c0090cca7');

    // Get Salt's "real time" event stream.
    source.onopen = function() { source.send('websocket client ready'); };

    // Other handlers
    source.onerror = function(e) { console.debug('error!', e); };

    // e.data represents Salt's "real time" event data as serialized JSON.
    source.onmessage = function(e) { console.debug(e.data); };

    // Terminates websocket connection and Salt's "real time" event stream on the server.
    source.close();

Or via Python, using the Python module
`websocket-client <https://pypi.python.org/pypi/websocket-client/>`_ for example.
Or the tornado
`client <https://tornado.readthedocs.io/en/latest/websocket.html#client-side-support>`_.

.. code-block:: python

    # Note, you must be authenticated!

    from websocket import create_connection

    # Get the Websocket connection to Salt
    ws = create_connection('wss://localhost:8000/formatted_events/d0ce6c1a37e99dcc0374392f272fe19c0090cca7')

    # Get Salt's "real time" event stream.
    ws.send('websocket client ready')


    # Simple listener to print results of Salt's "real time" event stream.
    # Look at https://pypi.python.org/pypi/websocket-client/ for more examples.
    while listening_to_events:
        print ws.recv()       #  Salt's "real time" event data as serialized JSON.

    # Terminates websocket connection and Salt's "real time" event stream on the server.
    ws.close()

    # Please refer to https://github.com/liris/websocket-client/issues/81 when using a self signed cert

Above examples show how to establish a websocket connection to Salt and activating
real time updates from Salt's event stream by signaling ``websocket client ready``.

Example responses
-----------------

``Minion information`` is a dictionary keyed by each connected minion's ``id`` (``mid``),
grains information for each minion is also included.

Minion information is sent in response to the following minion events:

- connection drops
    - requires running ``manage.present`` periodically every ``loop_interval`` seconds
- minion addition
- minion removal

.. code-block:: python

    # Not all grains are shown
    data: {
        "minions": {
            "minion1": {
                "id": "minion1",
                "grains": {
                    "kernel": "Darwin",
                    "domain": "local",
                    "zmqversion": "4.0.3",
                    "kernelrelease": "13.2.0"
                }
            }
        }
    }

``Job information`` is also tracked and delivered.

Job information is also a dictionary
in which each job's information is keyed by salt's ``jid``.

.. code-block:: python

    data: {
        "jobs": {
            "20140609153646699137": {
                "tgt_type": "glob",
                "jid": "20140609153646699137",
                "tgt": "*",
                "start_time": "2014-06-09T15:36:46.700315",
                "state": "complete",
                "fun": "test.ping",
                "minions": {
                    "minion1": {
                        "return": true,
                        "retcode": 0,
                        "success": true
                    }
                }
            }
        }
    }

Setup
=====
"""

import logging

import salt.ext.tornado.gen
import salt.ext.tornado.websocket
import salt.netapi
import salt.utils.json

from . import event_processor
from .saltnado import _check_cors_origin

_json = salt.utils.json.import_json()


log = logging.getLogger(__name__)


class AllEventsHandler(
    salt.ext.tornado.websocket.WebSocketHandler
):  # pylint: disable=W0223,W0232
    """
    Server side websocket handler.
    """

    # pylint: disable=W0221
    def get(self, token):
        """
        Check the token, returns a 401 if the token is invalid.
        Else open the websocket connection
        """
        log.debug("In the websocket get method")

        self.token = token
        # close the connection, if not authenticated
        if not self.application.auth.get_tok(token):
            log.debug("Refusing websocket connection, bad token!")
            self.send_error(401)
            return
        super().get(token)

    def open(self, token):  # pylint: disable=W0221
        """
        Return a websocket connection to Salt
        representing Salt's "real time" event stream.
        """
        self.connected = False

    @salt.ext.tornado.gen.coroutine
    def on_message(self, message):
        """Listens for a "websocket client ready" message.
        Once that message is received an asynchronous job
        is stated that yields messages to the client.
        These messages make up salt's
        "real time" event stream.
        """
        log.debug("Got websocket message %s", message)
        if message == "websocket client ready":
            if self.connected:
                # TBD: Add ability to run commands in this branch
                log.debug("Websocket already connected, returning")
                return

            self.connected = True

            while True:
                try:
                    event = yield self.application.event_listener.get_event(self)
                    self.write_message(salt.utils.json.dumps(event, _json_module=_json))
                except Exception as err:  # pylint: disable=broad-except
                    log.info(
                        "Error! Ending server side websocket connection. Reason = %s",
                        err,
                    )
                    break

            self.close()
        else:
            # TBD: Add logic to run salt commands here
            pass

    def on_close(self, *args, **kwargs):
        """Cleanup."""
        log.debug("In the websocket close method")
        self.close()

    def check_origin(self, origin):
        """
        If cors is enabled, check that the origin is allowed
        """

        mod_opts = self.application.mod_opts

        if mod_opts.get("cors_origin"):
            return bool(_check_cors_origin(origin, mod_opts["cors_origin"]))
        else:
            return super().check_origin(origin)


class FormattedEventsHandler(AllEventsHandler):  # pylint: disable=W0223,W0232
    @salt.ext.tornado.gen.coroutine
    def on_message(self, message):
        """Listens for a "websocket client ready" message.
        Once that message is received an asynchronous job
        is stated that yields messages to the client.
        These messages make up salt's
        "real time" event stream.
        """
        log.debug("Got websocket message %s", message)
        if message == "websocket client ready":
            if self.connected:
                # TBD: Add ability to run commands in this branch
                log.debug("Websocket already connected, returning")
                return

            self.connected = True

            evt_processor = event_processor.SaltInfo(self)
            client = salt.netapi.NetapiClient(self.application.opts)
            client.run(
                {
                    "fun": "grains.items",
                    "tgt": "*",
                    "token": self.token,
                    "mode": "client",
                    "asynchronous": "local_async",
                    "client": "local",
                }
            )
            while True:
                try:
                    event = yield self.application.event_listener.get_event(self)
                    evt_processor.process(event, self.token, self.application.opts)
                    # self.write_message('data: {0}\n\n'.format(salt.utils.json.dumps(event, _json_module=_json)))
                except Exception as err:  # pylint: disable=broad-except
                    log.debug(
                        "Error! Ending server side websocket connection. Reason = %s",
                        err,
                    )
                    break

            self.close()
        else:
            # TBD: Add logic to run salt commands here
            pass



"""
Test process name behavior with multiprocessing enabled and disabled.
"""


def test_process_name_no_pollution_when_multiprocessing_disabled(
    salt_master_factory,
):
    """
    Test that process names don't accumulate _thread_return strings
    when multiprocessing is disabled (issue #68553).

    When multiprocessing=False, jobs run in threads in the main process.
    This test ensures that the process name doesn't keep appending
    "Minion._thread_return" with each job execution.
    """
    # Create a fresh master for this test
    master = salt_master_factory.salt_master_daemon("test-process-name-master")

    # Create a fresh minion with multiprocessing=False
    minion = master.salt_minion_daemon(
        "test-process-name-minion-no-mp",
        overrides={"multiprocessing": False},
    )

    cli = master.salt_cli()

    with master.started(), minion.started():
        # Execute multiple jobs to ensure the process name doesn't accumulate
        for i in range(3):
            ret = cli.run("test.ping", minion_tgt=minion.id)
            assert ret.returncode == 0
            assert ret.data is True

        # Get the minion's PID directly from the fixture
        minion_pid = minion.pid
        assert minion_pid, "Minion PID not available"

        # Get the process info using the actual PID
        ret = cli.run("ps.proc_info", minion_pid, minion_tgt=minion.id)
        assert ret.returncode == 0, f"Failed to get process info for PID {minion_pid}"

        # Get the process command line
        proc_info = ret.data
        cmdline = " ".join(proc_info.get("cmdline", []))

        # The process name should not contain _thread_return at all
        # when multiprocessing=False, as we skip the appendproctitle call
        assert "_thread_return" not in cmdline, (
            f"Process cmdline should not contain '_thread_return' when "
            f"multiprocessing=False, but got: {cmdline}"
        )

        # Verify the process is a salt-minion process
        assert (
            "salt-minion" in cmdline
            or "salt_minion" in cmdline
            or "cli_salt_minion" in cmdline
        )


def test_process_name_normal_when_multiprocessing_enabled(
    salt_master_factory,
):
    """
    Test that process names work normally when multiprocessing is enabled.

    When multiprocessing=True (default), jobs run in separate processes,
    so the main minion process name should remain clean.
    """
    # Create a fresh master for this test
    master = salt_master_factory.salt_master_daemon("test-process-name-master-mp")

    # Create a fresh minion with multiprocessing=True
    minion = master.salt_minion_daemon(
        "test-process-name-minion-with-mp",
        overrides={"multiprocessing": True},
    )

    cli = master.salt_cli()

    with master.started(), minion.started():
        # Execute a job
        ret = cli.run("test.ping", minion_tgt=minion.id)
        assert ret.returncode == 0
        assert ret.data is True

        # Get the minion's PID directly from the fixture
        minion_pid = minion.pid
        assert minion_pid, "Minion PID not available"

        # Get the process info using the actual PID
        ret = cli.run("ps.proc_info", minion_pid, minion_tgt=minion.id)
        assert ret.returncode == 0, f"Failed to get process info for PID {minion_pid}"

        # Get the process command line
        proc_info = ret.data
        cmdline = " ".join(proc_info.get("cmdline", []))

        # Verify the process is a salt-minion process
        assert (
            "salt-minion" in cmdline
            or "salt_minion" in cmdline
            or "cli_salt_minion" in cmdline
        )

        # The main minion process should NOT accumulate _thread_return when multiprocessing=True
        # because jobs run in separate child processes
        assert "_thread_return" not in cmdline, (
            f"Main process cmdline should not contain '_thread_return' when "
            f"multiprocessing=True, but got: {cmdline}"
        )


def test_process_name_includes_minion_process_manager(
    salt_master_factory,
):
    """
    Test that the process name includes MinionProcessManager.

    This verifies that minion process managers append their name to the
    process title even when running in MainProcess (e.g., with --disable-keepalive).
    """
    # Create a fresh master for this test
    master = salt_master_factory.salt_master_daemon("test-process-name-master-pm")

    # Create a fresh minion
    minion = master.salt_minion_daemon(
        "test-process-name-minion-pm",
    )

    cli = master.salt_cli()

    with master.started(), minion.started():
        # Execute a job to ensure minion is fully running
        ret = cli.run("test.ping", minion_tgt=minion.id)
        assert ret.returncode == 0
        assert ret.data is True

        # Get the minion's PID directly from the fixture
        minion_pid = minion.pid
        assert minion_pid, "Minion PID not available"

        # Get the process info using the actual PID
        ret = cli.run("ps.proc_info", minion_pid, minion_tgt=minion.id)
        assert ret.returncode == 0, f"Failed to get process info for PID {minion_pid}"

        # Get the process command line
        proc_info = ret.data
        cmdline = " ".join(proc_info.get("cmdline", []))

        # The process title should include either MinionProcessManager or MultiMinionProcessManager
        # This validates the fix for minion process managers to append their name
        # even when running in MainProcess
        has_minion_pm = (
            "MinionProcessManager" in cmdline or "MultiMinionProcessManager" in cmdline
        )
        assert has_minion_pm, (
            f"Process cmdline should contain 'MinionProcessManager' or "
            f"'MultiMinionProcessManager', but got: {cmdline}"
        )



"""
    :codeauthor: Rahul Handay <rahulha@saltstack.com>

    Test cases for salt.modules.pagerduty
"""

import pytest

import salt.modules.pagerduty as pagerduty
import salt.utils.json
import salt.utils.pagerduty
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {pagerduty: {"__salt__": {"config.option": MagicMock(return_value=None)}}}


def test_list_services():
    """
    Test for List services belonging to this account
    """
    with patch.object(salt.utils.pagerduty, "list_items", return_value="A"):
        assert pagerduty.list_services() == "A"


def test_list_incidents():
    """
    Test for List incidents belonging to this account
    """
    with patch.object(salt.utils.pagerduty, "list_items", return_value="A"):
        assert pagerduty.list_incidents() == "A"


def test_list_users():
    """
    Test for List users belonging to this account
    """
    with patch.object(salt.utils.pagerduty, "list_items", return_value="A"):
        assert pagerduty.list_users() == "A"


def test_list_schedules():
    """
    Test for List schedules belonging to this account
    """
    with patch.object(salt.utils.pagerduty, "list_items", return_value="A"):
        assert pagerduty.list_schedules() == "A"


def test_list_windows():
    """
    Test for List maintenance windows belonging to this account
    """
    with patch.object(salt.utils.pagerduty, "list_items", return_value="A"):
        assert pagerduty.list_windows() == "A"


def test_list_policies():
    """
    Test for List escalation policies belonging to this account
    """
    with patch.object(salt.utils.pagerduty, "list_items", return_value="A"):
        assert pagerduty.list_policies() == "A"


def test_create_event():
    """
    Test for Create an event in PagerDuty. Designed for use in states.
    """
    with patch.object(salt.utils.json, "loads", return_value=["A"]):
        with patch.object(salt.utils.pagerduty, "query", return_value="A"):
            assert pagerduty.create_event() == ["A"]



import pytest

import salt.states.win_wusa as wusa
from salt.exceptions import SaltInvocationError
from tests.support.mock import MagicMock, patch


@pytest.fixture
def kb():
    return "KB123456"


@pytest.fixture
def configure_loader_modules():
    return {wusa: {"__opts__": {"test": False}, "__env__": "base"}}


def test_installed_no_source():
    """
    test wusa.installed without passing source
    """
    with pytest.raises(SaltInvocationError) as excinfo:
        wusa.installed(name="KB123456", source=None)
        assert excinfo.exception.strerror == 'Must specify a "source" file to install'


def test_installed_existing(kb):
    """
    test wusa.installed when the kb is already installed
    """
    mock_installed = MagicMock(return_value=True)
    with patch.dict(wusa.__salt__, {"wusa.is_installed": mock_installed}):
        returned = wusa.installed(name=kb, source=f"salt://{kb}.msu")
        expected = {
            "changes": {},
            "comment": f"{kb} already installed",
            "name": kb,
            "result": True,
        }
        assert expected == returned


def test_installed_test_true(kb):
    """
    test wusa.installed with test=True
    """
    mock_installed = MagicMock(return_value=False)
    with patch.dict(wusa.__salt__, {"wusa.is_installed": mock_installed}), patch.dict(
        wusa.__opts__, {"test": True}
    ):
        returned = wusa.installed(name=kb, source=f"salt://{kb}.msu")
        expected = {
            "changes": {},
            "comment": f"{kb} would be installed",
            "name": kb,
            "result": None,
        }
        assert expected == returned


def test_installed_cache_fail(kb):
    """
    test wusa.install when it fails to cache the file
    """
    mock_installed = MagicMock(return_value=False)
    mock_cache = MagicMock(return_value="")
    with patch.dict(
        wusa.__salt__,
        {"wusa.is_installed": mock_installed, "cp.cache_file": mock_cache},
    ):
        returned = wusa.installed(name=kb, source=f"salt://{kb}.msu")
        expected = {
            "changes": {},
            "comment": f'Unable to cache salt://{kb}.msu from saltenv "base"',
            "name": kb,
            "result": False,
        }
        assert expected == returned


def test_installed(kb):
    """
    test wusa.installed assuming success
    """
    mock_installed = MagicMock(side_effect=[False, True])
    mock_cache = MagicMock(return_value=f"C:\\{kb}.msu")
    with patch.dict(
        wusa.__salt__,
        {
            "wusa.is_installed": mock_installed,
            "cp.cache_file": mock_cache,
            "wusa.install": MagicMock(),
        },
    ):
        returned = wusa.installed(name=kb, source=f"salt://{kb}.msu")
        expected = {
            "changes": {"new": True, "old": False},
            "comment": f"{kb} was installed. ",
            "name": kb,
            "result": True,
        }
        assert expected == returned


def test_installed_failed(kb):
    """
    test wusa.installed with a failure
    """
    mock_installed = MagicMock(side_effect=[False, False])
    mock_cache = MagicMock(return_value=f"C:\\{kb}.msu")
    with patch.dict(
        wusa.__salt__,
        {
            "wusa.is_installed": mock_installed,
            "cp.cache_file": mock_cache,
            "wusa.install": MagicMock(),
        },
    ):
        returned = wusa.installed(name=kb, source=f"salt://{kb}.msu")
        expected = {
            "changes": {},
            "comment": f"{kb} failed to install. ",
            "name": kb,
            "result": False,
        }
        assert expected == returned


def test_uninstalled_non_existing(kb):
    """
    test wusa.uninstalled when the kb is not installed
    """
    mock_installed = MagicMock(return_value=False)
    with patch.dict(wusa.__salt__, {"wusa.is_installed": mock_installed}):
        returned = wusa.uninstalled(name=kb)
        expected = {
            "changes": {},
            "comment": f"{kb} already uninstalled",
            "name": kb,
            "result": True,
        }
        assert expected == returned


def test_uninstalled_test_true(kb):
    """
    test wusa.uninstalled with test=True
    """
    mock_installed = MagicMock(return_value=True)
    with patch.dict(wusa.__salt__, {"wusa.is_installed": mock_installed}), patch.dict(
        wusa.__opts__, {"test": True}
    ):
        returned = wusa.uninstalled(name=kb)
        expected = {
            "changes": {},
            "comment": f"{kb} would be uninstalled",
            "name": kb,
            "result": None,
        }
        assert expected == returned


def test_uninstalled(kb):
    """
    test wusa.uninstalled assuming success
    """
    mock_installed = MagicMock(side_effect=[True, False])
    with patch.dict(
        wusa.__salt__,
        {"wusa.is_installed": mock_installed, "wusa.uninstall": MagicMock()},
    ):
        returned = wusa.uninstalled(name=kb)
        expected = {
            "changes": {"new": False, "old": True},
            "comment": f"{kb} was uninstalled",
            "name": kb,
            "result": True,
        }
        assert expected == returned


def test_uninstalled_failed(kb):
    """
    test wusa.uninstalled with a failure
    """
    mock_installed = MagicMock(side_effect=[True, True])
    with patch.dict(
        wusa.__salt__,
        {"wusa.is_installed": mock_installed, "wusa.uninstall": MagicMock()},
    ):
        returned = wusa.uninstalled(name=kb)
        expected = {
            "changes": {},
            "comment": f"{kb} failed to uninstall",
            "name": kb,
            "result": False,
        }
        assert expected == returned



import pytest

import salt.loader.context
import salt.modules.state as statemod
import salt.modules.transactional_update as tu
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, patch

pytestmark = [
    pytest.mark.skip_on_windows(reason="Not supported on Windows"),
]


@pytest.fixture
def configure_loader_modules():
    loader_context = salt.loader.context.LoaderContext()
    return {
        tu: {
            "__salt__": {},
            "__utils__": {"files.rm_rf": MagicMock()},
            "__pillar__": salt.loader.context.NamedLoaderContext(
                "__pillar__", loader_context, {}
            ),
            "__opts__": {"extension_modules": "", "cachedir": "/tmp/"},
        },
        statemod: {"__salt__": {}, "__context__": {}},
    }


def test__global_params_no_self_update():
    """Test transactional_update._global_params without self_update"""
    assert tu._global_params(self_update=False) == [
        "--non-interactive",
        "--drop-if-no-change",
        "--no-selfupdate",
    ]


def test__global_params_self_update():
    """Test transactional_update._global_params with self_update"""
    assert tu._global_params(self_update=True) == [
        "--non-interactive",
        "--drop-if-no-change",
    ]


def test__global_params_no_self_update_snapshot():
    """Test transactional_update._global_params without self_update and
    snapshot

    """
    assert tu._global_params(self_update=False, snapshot=10) == [
        "--non-interactive",
        "--drop-if-no-change",
        "--no-selfupdate",
        "--continue",
        10,
    ]


def test__global_params_no_self_update_continue():
    """Test transactional_update._global_params without self_update and
    snapshot conitue

    """
    assert tu._global_params(self_update=False, snapshot="continue") == [
        "--non-interactive",
        "--drop-if-no-change",
        "--no-selfupdate",
        "--continue",
    ]


def test__pkg_params_no_packages():
    """Test transactional_update._pkg_params without packages"""
    with pytest.raises(CommandExecutionError):
        tu._pkg_params(pkg=None, pkgs=None, args=None)


def test__pkg_params_pkg():
    """Test transactional_update._pkg_params with single package"""
    assert tu._pkg_params(pkg="pkg1", pkgs=None, args=None) == ["pkg1"]


def test__pkg_params_pkgs():
    """Test transactional_update._pkg_params with packages"""
    assert tu._pkg_params(pkg=None, pkgs="pkg1", args=None) == ["pkg1"]
    assert tu._pkg_params(pkg=None, pkgs="pkg1 pkg2 ", args=None) == [
        "pkg1",
        "pkg2",
    ]
    assert tu._pkg_params(pkg=None, pkgs=["pkg1", "pkg2"], args=None) == [
        "pkg1",
        "pkg2",
    ]


def test__pkg_params_pkg_pkgs():
    """Test transactional_update._pkg_params with packages"""
    assert tu._pkg_params(pkg="pkg1", pkgs="pkg2", args=None) == [
        "pkg1",
        "pkg2",
    ]


def test__pkg_params_args():
    """Test transactional_update._pkg_params with argumens"""
    assert tu._pkg_params(pkg="pkg1", pkgs=None, args="--arg1") == [
        "--arg1",
        "pkg1",
    ]
    assert tu._pkg_params(pkg="pkg1", pkgs=None, args="--arg1 --arg2") == [
        "--arg1",
        "--arg2",
        "pkg1",
    ]
    assert tu._pkg_params(pkg="pkg1", pkgs=None, args=["--arg1", "--arg2"]) == [
        "--arg1",
        "--arg2",
        "pkg1",
    ]


def test_transactional_transactional():
    """Test transactional_update.transactional"""
    matrix = (("/usr/sbin/transactional-update", True), ("", False))

    for path_which, result in matrix:
        utils_mock = {"path.which": MagicMock(return_value=path_which)}

        with patch.dict(tu.__utils__, utils_mock):
            assert tu.transactional() is result
            utils_mock["path.which"].assert_called_with("transactional-update")


def test_in_transaction():
    """Test transactional_update.in_transaction"""
    matrix = (
        ("/usr/sbin/transactional-update", True, True),
        ("/usr/sbin/transactional-update", False, False),
        ("", True, False),
        ("", False, False),
    )

    for path_which, in_chroot, result in matrix:
        utils_mock = {"path.which": MagicMock(return_value=path_which)}
        salt_mock = {"chroot.in_chroot": MagicMock(return_value=in_chroot)}

        with patch.dict(tu.__utils__, utils_mock):
            with patch.dict(tu.__salt__, salt_mock):
                assert tu.in_transaction() is result


def test_commands_with_global_params():
    """Test commands that only accept global params"""
    for cmd in [
        "cleanup",
        "cleanup_snapshots",
        "cleanup_overlays",
        "grub_cfg",
        "bootloader",
        "initrd",
        "kdump",
        "reboot",
        "dup",
        "up",
        "patch",
        "migration",
    ]:
        salt_mock = {
            "cmd.run_all": MagicMock(return_value={"stdout": "output", "retcode": 0})
        }
        with patch.dict(tu.__salt__, salt_mock):
            assert getattr(tu, cmd)() == "output"
            salt_mock["cmd.run_all"].assert_called_with(
                [
                    "transactional-update",
                    "--non-interactive",
                    "--drop-if-no-change",
                    "--no-selfupdate",
                    (
                        cmd.replace("_", ".")
                        if cmd.startswith("grub")
                        else cmd.replace("_", "-")
                    ),
                ]
            )


def test_run_error():
    """Test transactional_update.run with missing command"""
    with pytest.raises(CommandExecutionError):
        tu.run(None)


def test_run_string():
    """Test transactional_update.run with command as string"""
    salt_mock = {
        "cmd.run_all": MagicMock(return_value={"stdout": "output", "retcode": 0})
    }
    with patch.dict(tu.__salt__, salt_mock):
        assert tu.run("cmd --flag p1 p2") == "output"
        salt_mock["cmd.run_all"].assert_called_with(
            [
                "transactional-update",
                "--non-interactive",
                "--drop-if-no-change",
                "--no-selfupdate",
                "--quiet",
                "run",
                "cmd",
                "--flag",
                "p1",
                "p2",
            ]
        )


def test_run_array():
    """Test transactional_update.run with command as array"""
    salt_mock = {
        "cmd.run_all": MagicMock(return_value={"stdout": "output", "retcode": 0})
    }
    with patch.dict(tu.__salt__, salt_mock):
        assert tu.run(["cmd", "--flag", "p1", "p2"]) == "output"
        salt_mock["cmd.run_all"].assert_called_with(
            [
                "transactional-update",
                "--non-interactive",
                "--drop-if-no-change",
                "--no-selfupdate",
                "--quiet",
                "run",
                "cmd",
                "--flag",
                "p1",
                "p2",
            ]
        )


def test_pkg_commands():
    """Test transactional_update.pkg_* commands"""
    for cmd in ["pkg_install", "pkg_remove", "pkg_update"]:
        salt_mock = {
            "cmd.run_all": MagicMock(return_value={"stdout": "output", "retcode": 0})
        }
        with patch.dict(tu.__salt__, salt_mock):
            assert getattr(tu, cmd)("pkg1", "pkg2 pkg3", "--arg") == "output"
            salt_mock["cmd.run_all"].assert_called_with(
                [
                    "transactional-update",
                    "--non-interactive",
                    "--drop-if-no-change",
                    "--no-selfupdate",
                    "pkg",
                    cmd.replace("pkg_", ""),
                    "--arg",
                    "pkg1",
                    "pkg2",
                    "pkg3",
                ]
            )


def test_rollback_error():
    """Test transactional_update.rollback with wrong snapshot"""
    with pytest.raises(CommandExecutionError):
        tu.rollback("error")


def test_rollback_default():
    """Test transactional_update.rollback with default snapshot"""
    salt_mock = {
        "cmd.run_all": MagicMock(return_value={"stdout": "output", "retcode": 0})
    }
    with patch.dict(tu.__salt__, salt_mock):
        assert tu.rollback() == "output"
        salt_mock["cmd.run_all"].assert_called_with(
            ["transactional-update", "rollback"]
        )


def test_rollback_snapshot_number():
    """Test transactional_update.rollback with numeric snapshot"""
    salt_mock = {
        "cmd.run_all": MagicMock(return_value={"stdout": "output", "retcode": 0})
    }
    with patch.dict(tu.__salt__, salt_mock):
        assert tu.rollback(10) == "output"
        salt_mock["cmd.run_all"].assert_called_with(
            ["transactional-update", "rollback", 10]
        )


def test_rollback_snapshot_str():
    """Test transactional_update.rollback with string snapshot"""
    salt_mock = {
        "cmd.run_all": MagicMock(return_value={"stdout": "output", "retcode": 0})
    }
    with patch.dict(tu.__salt__, salt_mock):
        assert tu.rollback("10") == "output"
        salt_mock["cmd.run_all"].assert_called_with(
            ["transactional-update", "rollback", "10"]
        )


def test_rollback_last():
    """Test transactional_update.rollback with last snapshot"""
    salt_mock = {
        "cmd.run_all": MagicMock(return_value={"stdout": "output", "retcode": 0})
    }
    with patch.dict(tu.__salt__, salt_mock):
        assert tu.rollback("last") == "output"
        salt_mock["cmd.run_all"].assert_called_with(
            ["transactional-update", "rollback", "last"]
        )


def test_pending_transaction():
    """Test transactional_update.pending_transaction"""
    matrix = (
        (False, ["1", "2+", "3-"], True),
        (False, ["1", "2-", "3+"], True),
        (False, ["1", "2", "3*"], False),
    )

    for in_transaction, snapshots, result in matrix:
        salt_mock = {
            "cmd.run_all": MagicMock(return_value={"stdout": snapshots, "retcode": 0})
        }

        tu_in_transaction = "salt.modules.transactional_update.in_transaction"
        with patch(tu_in_transaction) as in_transaction_mock:
            in_transaction_mock.return_value = in_transaction
            with patch.dict(tu.__salt__, salt_mock):
                assert tu.pending_transaction() is result
                salt_mock["cmd.run_all"].assert_called_with(
                    ["snapper", "--no-dbus", "list", "--columns", "number"]
                )


def test_pending_transaction_in_transaction():
    """Test transactional_update.pending_transaction when in transaction"""
    tu_in_transaction = "salt.modules.transactional_update.in_transaction"
    with patch(tu_in_transaction) as in_transaction_mock:
        in_transaction_mock.return_value = True
        with pytest.raises(CommandExecutionError):
            tu.pending_transaction()


def test_call_fails_input_validation():
    """Test transactional_update.call missing function name"""
    with pytest.raises(CommandExecutionError):
        tu.call("")


def test_call_fails_function():
    """Test transactional_update.chroot when fails the function"""
    utils_mock = {
        "json.find_json": MagicMock(side_effect=ValueError()),
    }
    salt_mock = {
        "cmd.run_all": MagicMock(
            return_value={"retcode": 0, "stdout": "Not found", "stderr": ""}
        ),
    }
    with patch.dict(tu.__utils__, utils_mock), patch.dict(tu.__salt__, salt_mock):
        assert tu.call("test.ping") == {
            "result": False,
            "retcode": 1,
            "comment": "Not found",
        }

        salt_mock["cmd.run_all"].assert_called_with(
            [
                "transactional-update",
                "--non-interactive",
                "--drop-if-no-change",
                "--no-selfupdate",
                "--continue",
                "--quiet",
                "run",
                "salt-call",
                "--out",
                "json",
                "-l",
                "quiet",
                "--no-return-event",
                "--",
                "test.ping",
            ]
        )


def test_call_success_no_reboot():
    """Test transactional_update.chroot when succeed"""
    utils_mock = {
        "json.find_json": MagicMock(return_value={"return": "result"}),
    }
    salt_mock = {
        "cmd.run_all": MagicMock(return_value={"retcode": 0, "stdout": ""}),
    }
    with patch.dict(tu.__utils__, utils_mock), patch.dict(tu.__salt__, salt_mock):
        assert tu.call("test.ping") == "result"

        salt_mock["cmd.run_all"].assert_called_with(
            [
                "transactional-update",
                "--non-interactive",
                "--drop-if-no-change",
                "--no-selfupdate",
                "--continue",
                "--quiet",
                "run",
                "salt-call",
                "--out",
                "json",
                "-l",
                "quiet",
                "--no-return-event",
                "--",
                "test.ping",
            ]
        )


def test_call_success_reboot():
    """Test transactional_update.chroot when succeed and reboot"""
    pending_transaction_mock = MagicMock(return_value=True)
    reboot_mock = MagicMock()
    utils_mock = {
        "json.find_json": MagicMock(return_value={"return": "result"}),
    }
    salt_mock = {
        "cmd.run_all": MagicMock(return_value={"retcode": 0, "stdout": ""}),
    }
    with patch.dict(tu.__utils__, utils_mock), patch.dict(
        tu.__salt__, salt_mock
    ), patch.dict(tu.__salt__, salt_mock), patch(
        "salt.modules.transactional_update.pending_transaction",
        pending_transaction_mock,
    ), patch(
        "salt.modules.transactional_update.reboot", reboot_mock
    ):
        assert (
            tu.call("transactional_update.dup", activate_transaction=True) == "result"
        )

        salt_mock["cmd.run_all"].assert_called_with(
            [
                "transactional-update",
                "--non-interactive",
                "--drop-if-no-change",
                "--no-selfupdate",
                "--continue",
                "--quiet",
                "run",
                "salt-call",
                "--out",
                "json",
                "-l",
                "quiet",
                "--no-return-event",
                "--",
                "transactional_update.dup",
            ]
        )
        pending_transaction_mock.assert_called_once()
        reboot_mock.assert_called_once()


def test_call_success_parameters():
    """Test transactional_update.chroot when succeed with parameters"""
    utils_mock = {
        "json.find_json": MagicMock(return_value={"return": "result"}),
    }
    salt_mock = {
        "cmd.run_all": MagicMock(return_value={"retcode": 0, "stdout": ""}),
    }
    with patch.dict(tu.__utils__, utils_mock), patch.dict(tu.__salt__, salt_mock):
        assert tu.call("module.function", key="value") == "result"

        salt_mock["cmd.run_all"].assert_called_with(
            [
                "transactional-update",
                "--non-interactive",
                "--drop-if-no-change",
                "--no-selfupdate",
                "--continue",
                "--quiet",
                "run",
                "salt-call",
                "--out",
                "json",
                "-l",
                "quiet",
                "--no-return-event",
                "--",
                "module.function",
                "key=value",
            ]
        )


def test_sls():
    """Test transactional_update.sls"""
    salt_mock = {
        "saltutil.is_running": MagicMock(return_value=[]),
    }
    with patch.dict(statemod.__salt__, salt_mock), patch(
        "salt.modules.transactional_update.call", MagicMock(return_value="result")
    ):
        assert tu.sls("module") == "result"


def test_sls_queue_true():
    """Test transactional_update.sls"""
    salt_mock = {
        "saltutil.is_running": MagicMock(
            side_effect=[
                [
                    {
                        "fun": "state.running",
                        "pid": "4126",
                        "jid": "20150325123407204096",
                    }
                ],
                [],
            ]
        ),
    }
    with patch.dict(statemod.__salt__, salt_mock), patch(
        "salt.modules.transactional_update.call", MagicMock(return_value="result")
    ):
        assert tu.sls("module", queue=True) == "result"


def test_sls_queue_false_failing():
    """Test transactional_update.sls"""
    salt_mock = {
        "saltutil.is_running": MagicMock(
            side_effect=[
                [
                    {
                        "fun": "state.running",
                        "pid": "4126",
                        "jid": "20150325123407204096",
                    }
                ],
                [],
            ]
        ),
    }
    with patch.dict(statemod.__salt__, salt_mock), patch(
        "salt.modules.transactional_update.call", MagicMock(return_value="result")
    ):
        assert tu.sls("module", queue=False) == [
            'The function "state.running" is running as PID 4126 and was started at 2015, Mar 25 12:34:07.204096 with jid 20150325123407204096'
        ]


def test_highstate():
    """Test transactional_update.highstage"""
    salt_mock = {
        "saltutil.is_running": MagicMock(return_value=[]),
    }
    with patch.dict(statemod.__salt__, salt_mock), patch(
        "salt.modules.transactional_update.call", MagicMock(return_value="result")
    ):
        assert tu.highstate() == "result"


def test_highstate_queue_true():
    """Test transactional_update.highstage"""
    salt_mock = {
        "saltutil.is_running": MagicMock(
            side_effect=[
                [
                    {
                        "fun": "state.running",
                        "pid": "4126",
                        "jid": "20150325123407204096",
                    }
                ],
                [],
            ]
        ),
    }
    with patch.dict(statemod.__salt__, salt_mock), patch(
        "salt.modules.transactional_update.call", MagicMock(return_value="result")
    ):
        assert tu.highstate(queue=True) == "result"


def test_highstate_queue_false_failing():
    """Test transactional_update.highstage"""
    salt_mock = {
        "saltutil.is_running": MagicMock(
            side_effect=[
                [
                    {
                        "fun": "state.running",
                        "pid": "4126",
                        "jid": "20150325123407204096",
                    }
                ],
                [],
            ]
        ),
    }
    with patch.dict(statemod.__salt__, salt_mock), patch(
        "salt.modules.transactional_update.call", MagicMock(return_value="result")
    ):
        assert tu.highstate(queue=False) == [
            'The function "state.running" is running as PID 4126 and was started at 2015, Mar 25 12:34:07.204096 with jid 20150325123407204096'
        ]


def test_single():
    """Test transactional_update.single"""
    salt_mock = {
        "saltutil.is_running": MagicMock(return_value=[]),
    }
    with patch.dict(statemod.__salt__, salt_mock), patch(
        "salt.modules.transactional_update.call", MagicMock(return_value="result")
    ):
        assert tu.single("pkg.installed", name="emacs") == "result"


def test_single_queue_false_failing():
    """Test transactional_update.single"""
    salt_mock = {
        "saltutil.is_running": MagicMock(
            side_effect=[
                [
                    {
                        "fun": "state.running",
                        "pid": "4126",
                        "jid": "20150325123407204096",
                    }
                ],
                [],
            ]
        ),
    }
    with patch.dict(statemod.__salt__, salt_mock), patch(
        "salt.modules.transactional_update.call", MagicMock(return_value="result")
    ):
        assert tu.single("pkg.installed", name="emacs", queue=False) == [
            'The function "state.running" is running as PID 4126 and was started at 2015, Mar 25 12:34:07.204096 with jid 20150325123407204096'
        ]


def test_single_queue_true():
    """Test transactional_update.single"""
    salt_mock = {
        "saltutil.is_running": MagicMock(
            side_effect=[
                [
                    {
                        "fun": "state.running",
                        "pid": "4126",
                        "jid": "20150325123407204096",
                    }
                ],
                [],
            ]
        ),
    }
    with patch.dict(statemod.__salt__, salt_mock), patch(
        "salt.modules.transactional_update.call", MagicMock(return_value="result")
    ):
        assert tu.single("pkg.installed", name="emacs", queue=True) == "result"
